import sys
from collections import OrderedDict
from itertools import repeat

import dynet as dy
import dynet_config
import numpy as np
from tqdm import tqdm

from .birnn import EmptyRNN, BiRNN, HighwayRNN, HierarchicalBiRNN
from .constants import TRAINERS, TRAINER_LEARNING_RATE_PARAM_NAMES, TRAINER_KWARGS, CategoricalParameter
from .mlp import MultilayerPerceptron
from .sub_model import SubModel
from ..classifier import Classifier
from ...config import Config, BIRNN, HIGHWAY_RNN, HIERARCHICAL_RNN
from ...model_util import MISSING_VALUE, remove_existing

BIRNN_TYPES = {BIRNN: BiRNN, HIGHWAY_RNN: HighwayRNN, HIERARCHICAL_RNN: HierarchicalBiRNN}

tqdm.monitor_interval = 0

try:
    print("[dynet] %s" % dy.__gitversion__, file=sys.stderr, flush=True)
except AttributeError:
    pass


class AxisModel:
    """
    Format-specific parameters that are part of the network
    """
    def __init__(self, axis, num_labels, config, model, birnn_type):
        args = config.hyperparams.specific[axis]
        self.birnn = birnn_type(config, args, model, save_path=("axes", axis, "birnn"),
                                copy_shared=args.copy_shared == [] or axis in (args.copy_shared or ()))
        self.mlp = MultilayerPerceptron(config, args, model, num_labels=num_labels, save_path=("axes", axis, "mlp"))


class NeuralNetwork(Classifier, SubModel):
    """
    Neural network to be used by the parser for action classification. Uses dense features.
    Keeps weights in constant-size matrices. Does not allow adding new features on-the-fly.
    Allows adding new labels on-the-fly, but requires pre-setting maximum number of labels.
    Expects features from DenseFeatureExtractor.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new untrained NN
        """
        Classifier.__init__(self, *args, **kwargs)
        SubModel.__init__(self)
        self.minibatch_size = self.config.args.minibatch_size
        self.loss = self.config.args.loss
        self.weight_decay = self.config.args.dynet_weight_decay
        self.empty_values = OrderedDict()  # string (feature param key) -> expression
        self.axes = OrderedDict()  # string (axis) -> AxisModel
        self.losses = []
        self.steps = 0
        self.trainer_type = self.trainer = self.value = self.birnn = None

        if self.config.args.use_bert:
            import torch
            from pytorch_pretrained_bert import BertTokenizer, BertModel
            import logging

            self.torch = torch
            if self.config.args.bert_multilingual is not None:
                assert "multilingual" in self.config.args.bert_model
            logging.basicConfig(level=logging.INFO)
            is_uncased_model = "uncased" in self.config.args.bert_model
            self.tokenizer = BertTokenizer.from_pretrained(self.config.args.bert_model, do_lower_case=is_uncased_model)
            self.bert_model = BertModel.from_pretrained(self.config.args.bert_model)
            self.bert_model.eval()
            self.bert_model.to('cuda')
            self.bert_layers_count = 24 if "large" in self.config.args.bert_model else 12
            self.bert_embedding_len = 1024 if "large" in self.config.args.bert_model else 768

            self.last_weights = ""
        else:
            self.torch = self.tokenizer = self.bert_model = self.bert_layers_count = self.bert_embedding_len = \
                self.last_weights = None

    @property
    def input_dim(self):
        return OrderedDict((a, m.mlp.input_dim) for a, m in self.axes.items())
    
    @property
    def birnn_type(self):
        return BIRNN_TYPES.get(self.model_type, EmptyRNN)

    def resize(self):
        for axis, labels in self.labels.items():
            if labels.size is not None:
                num_labels = self.num_labels[axis]
                assert num_labels <= labels.size, "Exceeded maximum number of labels at axis '%s': %d > %d:\n%s" % (
                    axis, num_labels, labels.size, "\n".join(map(str, labels.all)))

    def init_model(self, axis=None, train=False):
        init = self.model is None
        if init:
            self.model = dy.ParameterCollection()
            self.birnn = self.birnn_type(self.config, Config().hyperparams.shared, self.model,
                                         save_path=("shared", "birnn"), shared=True)
            self.set_weight_decay_lambda()
        if train:
            self.init_trainer()
        if axis:
            self.init_axis_model(axis)
        if init:
            self.init_cg()
            self.finished_step()

    def set_weight_decay_lambda(self, weight_decay=None):
        self.model.set_weight_decay_lambda(self.weight_decay if weight_decay is None else weight_decay)

    def init_trainer(self):
        if self.trainer_type is None or str(self.trainer_type) != self.config.args.optimizer:
            self.trainer_type = CategoricalParameter(TRAINERS, self.config.args.optimizer)
            trainer_kwargs = dict(TRAINER_KWARGS.get(str(self.trainer_type), {}))
            learning_rate_param_name = TRAINER_LEARNING_RATE_PARAM_NAMES.get(str(self.trainer_type))
            if learning_rate_param_name and self.learning_rate:
                trainer_kwargs[learning_rate_param_name] = self.learning_rate
            self.config.print("Initializing trainer=%s(%s)" % (
                self.trainer_type, ", ".join("%s=%s" % (k, v) for k, v in trainer_kwargs.items())), level=4)
            self.trainer = self.trainer_type()(self.model, **trainer_kwargs)
            self.trainer.set_sparse_updates(False)

    def init_axis_model(self, axis, init=True):
        if axis in self.axes:
            if init:
                return
        else:
            self.axes[axis] = AxisModel(axis, self.labels[axis].size, self.config, self.model, self.birnn_type)
            self.config.print("Initializing %s model with %d labels" % (axis, self.labels[axis].size), level=4)
        indexed_dim = np.array([0, 0], dtype=int)  # specific, shared
        indexed_num = np.array([0, 0], dtype=int)
        for key, param in sorted(self.input_params.items()):
            if not param.enabled:
                continue
            if self.config.args.use_bert and \
                    (not self.config.args.bert_use_default_word_embeddings
                     or self.config.args.bert_multilingual is not None) \
                    and key == 'W':
                i = self.birnn_indices(param)
                indexed_num[i] = np.fmax(indexed_num[i], param.num)  # indices to be looked up are collected
                continue
            if self.config.args.bert_multilingual is not None and param.lang_specific:
                continue
            self.config.print("Initializing input parameter: %s" % param, level=4)
            if not param.numeric and key not in self.params:  # lookup feature
                if init:
                    lookup = self.model.add_lookup_parameters((param.size, param.dim))
                    lookup.set_updated(param.updated)
                    param.init_data()
                    if param.init is not None and param.init.size:
                        lookup.init_from_array(param.init)
                    self.params[key] = lookup
            if param.indexed:
                i = self.birnn_indices(param)
                indexed_dim[i] += param.dim  # add to the input dimensionality at each indexed time point
                indexed_num[i] = np.fmax(indexed_num[i], param.num)  # indices to be looked up are collected

        if self.config.args.use_bert and init:
            if self.config.args.bert_layers_pooling == "weighted":
                bert_weights = self.model.add_parameters(len(self.config.args.bert_layers), init=1)
                self.params["bert_weights"] = bert_weights

            indexed_dim[[0, 1]] += self.bert_embedding_len
            if self.config.args.bert_multilingual == 0:
                indexed_dim[[0, 1]] += 50

        for birnn in self.get_birnns(axis):
            birnn.init_params(indexed_dim[int(birnn.shared)], indexed_num[int(birnn.shared)])

    def birnn_indices(self, param):  # both specific and shared or just specific
        return [0, 1] if not self.config.args.multilingual or not param.lang_specific else [0]

    def init_cg(self, renew=True):
        if renew:
            dy.renew_cg()
        self.empty_values.clear()

    def get_empty_values(self, key):
        value = self.empty_values.get(key)
        if value is None:
            self.empty_values[key] = value = dy.inputVector(np.zeros(self.input_params[key].dim, dtype=float))
        return value

    def get_bert_embed(self, passage, lang, train=False):
        orig_tokens = passage
        bert_tokens = []
        # Token map will be an int -> int mapping between the `orig_tokens` index and
        # the `bert_tokens` index.
        orig_to_tok_map = []

        # Example:
        # orig_tokens = ["John", "Johanson", "'s",  "house"]
        # bert_tokens == ["[CLS]", "john", "johan", "##son", "'", "s", "house", "[SEP]"]
        # orig_to_tok_map == [(1), (2,3), (4,5), (6)]

        bert_tokens.append("[CLS]")
        for orig_token in orig_tokens:
            start_token = len(bert_tokens)
            bert_token = self.tokenizer.tokenize(orig_token)
            bert_tokens.extend(bert_token)
            end_token = start_token + len(bert_token)
            orig_to_tok_map.append(slice(start_token, end_token))
        bert_tokens.append("[SEP]")

        indexed_tokens = self.tokenizer.convert_tokens_to_ids(bert_tokens)
        tokens_tensor = self.torch.tensor([indexed_tokens])
        tokens_tensor = tokens_tensor.to('cuda')

        with self.torch.no_grad():
            encoded_layers, _ = self.bert_model(tokens_tensor)
        assert len(encoded_layers) == self.bert_layers_count, "Invalid BERT layer count %s" % len(encoded_layers)

        aligned_layer = []
        for layer in range(self.bert_layers_count):
            aligned_layer.append([])
            for mapping_range in orig_to_tok_map:
                token_embeddings = encoded_layers[layer][0][mapping_range]
                if self.config.args.bert_token_align_by == "mean":
                    aligned_layer[layer].append(self.torch.mean(token_embeddings, dim=(0,)).cpu().data.numpy())
                elif self.config.args.bert_token_align_by == "sum":
                    aligned_layer[layer].append(self.torch.sum(token_embeddings, dim=(0,)).cpu().data.numpy())
                elif self.config.args.bert_token_align_by == "first":
                    aligned_layer[layer].append(token_embeddings[0].cpu().data.numpy())
                else:
                    raise ValueError("Invalid BERT token align option '%s'" % self.config.args.bert_token_align_by)

        layer_list_to_use = self.config.args.bert_layers
        aligned_layer = [aligned_layer[i] for i in layer_list_to_use]

        if self.config.args.bert_layers_pooling == "weighted":
            bert_softmax = dy.softmax(self.params["bert_weights"])
            embeds = dy.cmult(dy.inputTensor(np.asarray(aligned_layer)), bert_softmax)
            embeds = dy.sum_dim(embeds, [0])
        elif self.config.args.bert_layers_pooling == "concat":
            embeds = dy.inputTensor(np.concatenate(aligned_layer, axis=1))
        elif self.config.args.bert_layers_pooling == "sum":
            embeds = dy.inputTensor(np.sum(aligned_layer, axis=0))
        else:
            raise ValueError("Invalid BERT pooling option '%s'" % self.config.args.bert_layers_pooling)

        if self.config.args.bert_multilingual == 0:
            assert lang
            if (lang + "_embed") in self.params:
                lang_embed = self.params[lang + "_embed"]
            else:
                lang_embed = self.model.add_parameters(50, init='glorot')
                self.params[lang + "_embed"] = lang_embed

            multilingual_embeds = []
            for embed in embeds:
                multilingual_embeds.append(dy.concatenate([lang_embed, embed]))

            embeds = dy.transpose(dy.concatenate_cols(multilingual_embeds))

        if self.config.args.bert_layers_pooling == "weighted":
            single_token_embed_len = self.bert_embedding_len
        elif self.config.args.bert_layers_pooling == "concat":
            single_token_embed_len = self.bert_embedding_len * len(layer_list_to_use)
        elif self.config.args.bert_layers_pooling == "sum":
            single_token_embed_len = self.bert_embedding_len
        else:
            raise ValueError("Invalid BERT pooling option '%s'" % self.config.args.bert_layers_pooling)
        if self.config.args.bert_multilingual == 0:
            single_token_embed_len += 50

        # TODO: try dropout strategies like dropping at the per layer embeddings or dropping entire layers.
        assert embeds.dim() == ((len(passage), single_token_embed_len), 1), "Invalid BERT dim %s" % embeds.dim()

        assert 0 <= self.config.args.bert_dropout < 1, "Invalid BERT dropout %s" % self.config.args.bert_dropout
        if train:
            embeds = dy.dropout(embeds, self.config.args.bert_dropout)

        return embeds

    def init_features(self, features, axes, train=False, passage=None, lang=None):
        for axis in axes:
            self.init_model(axis, train)
        embeddings = [[], []]  # specific, shared
        self.config.print("Initializing %s %s features for %d elements" %
                          (", ".join(axes), self.birnn_type.__name__, len(features)), level=4)
        for key, indices in sorted(features.items()):
            param = self.input_params[key]
            lookup = self.params.get(key)
            if not param.indexed or lookup is None:
                continue
            vectors = [lookup[k] for k in indices]
            for index in self.birnn_indices(param):
                embeddings[index].append((key, vectors))
            self.config.print(lambda: "%s: %s" % (key, ", ".join("%d->%s" % (i, e.npvalue().tolist())
                                                                 for i, e in zip(indices, vectors))), level=4)
        if self.config.args.use_bert:
            bert_emded = self.get_bert_embed(passage, lang, train)
            embeddings[0].append(('BERT', bert_emded))
            embeddings[1].append(('BERT', bert_emded))

            if "bert_weights" in self.params and self.config.args.verbose >= 4:
                print("\n--Bert Weights Changed--: ")
                print(str(dy.softmax(self.params["bert_weights"]).value()) != self.last_weights)
                print("\n--Bert Weights--: ")
                print(str(dy.softmax(self.params["bert_weights"]).value()))
                self.last_weights = str(dy.softmax(self.params["bert_weights"]).value())

        for birnn in self.get_birnns(*axes):
            birnn.init_features(embeddings[int(birnn.shared)], train)

    def generate_inputs(self, features, axis):
        indices = []  # list, not set, in order to maintain consistent order
        for key, values in sorted(features.items()):
            param = self.input_params[key]
            lookup = self.params.get(key)
            if self.config.args.bert_multilingual is not None and param.lang_specific and key != 'W':
                continue
            if param.numeric:
                yield key, dy.inputVector(values)
            elif param.indexed:  # collect indices to be looked up
                indices += values  # DenseFeatureExtractor collapsed features so there are no repetitions between them
            elif lookup is None:  # ignored
                continue
            else:  # lookup feature
                yield from ((key, self.get_empty_values(key) if x == MISSING_VALUE else lookup[x]) for x in values)
            self.config.print(lambda: "%s: %s" % (key, values), level=4)
        if indices:
            for birnn in self.get_birnns(axis):
                yield from birnn.evaluate(indices)

    def get_birnns(self, *axes):
        """ Return shared + axis-specific BiRNNs """
        return [m.birnn for m in [self] + [self.axes[axis] for axis in axes]]

    def evaluate(self, features, axis, train=False):
        """
        Apply MLP and log softmax to input features
        :param features: dictionary of key, values for each feature type
        :param axis: axis of the label we are predicting
        :param train: whether to apply dropout
        :return: expression corresponding to log softmax applied to MLP output
        """
        self.init_model(axis, train)
        value = self.value.get(axis)
        if value is None:
            self.value[axis] = value = self.axes[axis].mlp.evaluate(self.generate_inputs(features, axis), train=train)
        return value

    def score(self, features, axis):
        """
        Calculate score for each label
        :param features: extracted feature values, of size input_size
        :param axis: axis of the label we are predicting
        :return: array with score for each label
        """
        super().score(features, axis)
        num_labels = self.num_labels[axis]
        if self.updates > 0 and num_labels > 1:
            if dynet_config.gpu():
                # RestrictedLogSoftmax is not implemented for GPU, so we move the value to CPU first
                value = dy.to_device(self.evaluate(features, axis), 'CPU')
                # then, we move it back to GPU (if the device name is '', the default device will be selected)
                value = dy.to_device(dy.log_softmax(value, restrict=list(range(num_labels))), '').npvalue()
            else:
                value = dy.log_softmax(self.evaluate(features, axis), restrict=list(range(num_labels))).npvalue()
            return value[:num_labels]
        self.config.print("  no updates done yet, returning zero vector.", level=4)
        return np.zeros(num_labels)

    def update(self, features, axis, pred, true, importance=None):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, in the form of a dict (name: value)
        :param axis: axis of the label we are predicting
        :param pred: label predicted by the classifier (non-negative integer bounded by num_labels[axis])
        :param true: true labels (non-negative integers bounded by num_labels[axis])
        :param importance: how much to scale the update for the weight update for each true label
        """
        super().update(features, axis, pred, true, importance)
        losses = self.calc_loss(self.evaluate(features, axis, train=True), axis, true, importance or repeat(1))
        self.config.print(lambda: "  loss=" + ", ".join("%g" % l.value() for l in losses), level=4)
        self.losses += losses
        self.steps += 1

    def calc_loss(self, scores, axis, true, importance):
        ret = [i * dy.pickneglogsoftmax(scores, t) for t, i in zip(true, importance)]
        if self.loss == "max_margin":
            ret.append(dy.max_dim(dy.log_softmax(scores, restrict=list(set(range(self.num_labels[axis])) - set(true)))))
        return ret

    def finished_step(self, train=False):
        super().invalidate_caches()

    def invalidate_caches(self):
        self.value = {}  # For caching the result of _evaluate

    def finished_item(self, train=False, renew=True):
        if self.steps >= self.minibatch_size:
            self.finalize()
        elif not train:
            self.init_cg(renew)
        self.finished_step(train)

    def transition(self, action, axis):
        for birnn in self.get_birnns(axis):
            birnn.transition(action)

    def finalize(self, finished_epoch=False, **kwargs):
        """
        Fit this model on collected samples
        :return self
        """
        super().finalize(finished_epoch=finished_epoch, **kwargs)
        assert self.model, "Cannot finalize a model without initializing it first"
        if self.losses:
            loss = dy.esum(self.losses)
            loss.forward()
            self.config.print(lambda: "Total loss from %d time steps: %g" % (self.steps, loss.value()), level=4)
            loss.backward()
            try:
                self.trainer.update()
            except RuntimeError as e:
                Config().log("Error in update(): %s\n" % e)
            self.init_cg()
            self.losses = []
            self.steps = 0
            self.updates += 1
        if finished_epoch:
            self.trainer.learning_rate /= (1 - self.learning_rate_decay)
        if self.config.args.verbose > 2:
            self.trainer.status()
        return self
            
    def sub_models(self):
        """ :return: ordered list of SubModels """
        axes = list(filter(None, map(self.axes.get, self.labels or self.labels_t)))
        return [self] + [m.mlp for m in axes] + [m.birnn for m in axes + [self]]
    
    def save_sub_model(self, d, *args):
        return SubModel.save_sub_model(
            self, d,
            ("loss", self.loss),
            ("weight_decay", self.weight_decay),
        )

    def load_sub_model(self, d, *args, **kwargs):
        d = SubModel.load_sub_model(self, d, *args, **kwargs)
        self.config.args.loss = self.loss = d["loss"]
        self.config.args.dynet_weight_decay = self.weight_decay = d.get("weight_decay",
                                                                        self.config.args.dynet_weight_decay)

    def save_model(self, filename, d):
        Classifier.save_model(self, filename, d)
        self.finalize()
        values = []
        for model in self.sub_models():
            values += model.save_sub_model(d)
            if self.config.args.verbose <= 3:
                self.config.print(model.params_str, level=1)
        self.config.print(self, level=1)
        self.save_param_values(filename, values)

    def save_param_values(self, filename, values):
        remove_existing(filename + ".data", filename + ".meta")
        try:
            self.set_weight_decay_lambda(0.0)  # Avoid applying weight decay due to clab/dynet#1206, we apply it on load
            dy.save(filename, tqdm(values, desc="Saving model to '%s'" % filename, unit="param", file=sys.stdout))
            self.set_weight_decay_lambda()
        except ValueError as e:
            print("Failed saving model: %s" % e)

    def load_model(self, filename, d):
        self.model = None
        self.init_model()
        values = self.load_param_values(filename, d)
        self.axes = OrderedDict()
        for axis, labels in self.labels_t.items():
            _, size = labels
            assert size, "Size limit for '%s' axis labels is %s" % (axis, size)
            self.axes[axis] = AxisModel(axis, size, self.config, self.model, self.birnn_type)
        for model in self.sub_models():
            model.load_sub_model(d, *values)
            del values[:len(model.params)]  # Take next len(model.params) values
            if self.config.args.verbose <= 3:
                self.config.print(model.params_str)
        self.copy_shared_birnn(filename, d)
        assert not values, "Loaded values: %d more than expected" % len(values)
        if self.weight_decay and self.config.args.dynet_apply_weight_decay_on_load:
            t = tqdm(list(self.all_params(as_array=False).items()),
                     desc="Applying weight decay of %g" % self.weight_decay, unit="param", file=sys.stdout)
            for key, param in t:
                t.set_postfix(param=key)
                try:
                    value = param.as_array() * np.float_power(1 - self.weight_decay, self.updates)
                except AttributeError:
                    continue
                try:
                    param.set_value(value)
                except AttributeError:
                    param.init_from_array(value)
        self.config.print(self, level=1)

    def load_param_values(self, filename, d=None):
        return list(tqdm(dy.load_generator(filename, self.model), total=self.params_num(d) if d else None,
                         desc="Loading model from '%s'" % filename, unit="param", file=sys.stdout))

    def copy_shared_birnn(self, filename, d):
        shared_values = None
        values = self.load_param_values(filename, d)  # Load parameter values again so that shared parameters are copied
        for model in self.sub_models():
            if model is self.birnn:
                shared_values = values[:len(model.params)]
            del values[:len(model.params)]  # Take next len(model.params) values
        for axis, model in self.axes.items():
            if model.birnn.copy_shared:
                model.birnn.load_sub_model(d, *shared_values, load_path=self.birnn.save_path)
                if self.config.args.verbose <= 3:
                    self.config.print(lambda: "Copied from %s to %s" %
                                      ("/".join(self.birnn.save_path), model.birnn.params_str()), level=1)
                self.init_axis_model(axis, init=False)  # Update input_dim

    def params_num(self, d):
        return sum(len(m.get_sub_dict(d).get("param_keys", ())) for m in self.sub_models())

    def all_params(self, as_array=True):
        d = super().all_params()
        for model in self.sub_models():
            for key, value in model.params.items():
                for name, param in ((key, value),) if isinstance(value, (dy.Parameters, dy.LookupParameters)) else [
                        ("%s%s%d%d%d" % (key, p, i, j, k), v) for i, (f, b) in
                        enumerate(value.builder_layers)
                        for p, r in (("f", f), ("b", b)) for j, l in enumerate(r.get_parameters())
                        for k, v in enumerate(l)] if isinstance(value, dy.BiRNNBuilder) else [
                        ("%s%d%d" % (key, j, k), v) for j, l in enumerate(value.get_parameters())
                        for k, v in enumerate(l)]:
                    d["_".join(model.save_path + (name,))] = param.as_array() if as_array else param
        return d

    def print_params(self, max_rows=10):
        for model in self.sub_models():
            for key, value in model.params.items():
                print("[%s] %s" % (model.params_str(), key))
                # noinspection PyBroadException
                try:
                    print(value.as_array()[:max_rows])
                except Exception:
                    pass



import faulthandler
import sys

import os
import signal


def tracefunc(frame, event, arg):
    if event.endswith("call") and arg:
        if (getattr(arg, "__module__", None) or getattr(arg.__self__.__class__, "__module__")) == "_dynet":
            print(">", os.path.basename(frame.f_code.co_filename), frame.f_code.co_name, arg.__qualname__, "(",
                  ", ".join("%s=%r" % (v, frame.f_locals[v])
                            for v in frame.f_code.co_varnames[:frame.f_code.co_argcount] if v != "self"), ")")
            # print(arg.__qualname__)
    return tracefunc


def set_tracefunc():
    sys.setprofile(tracefunc)


def print_traceback(*args, **kwargs):
    del args, kwargs
    faulthandler.dump_traceback()


def set_traceback_listener(sig=None):
    try:
        signal.signal(signal.SIGUSR1 if sig is None else sig, print_traceback)
    except AttributeError:
        pass



from ..config import Config


class Edge:
    """
    Temporary representation for core.Edge with only relevant information for parsing
    """
    def __init__(self, parent, child, tag, remote=False):
        self.parent = parent  # Node object from which this edge comes
        self.child = child  # Node object to which this edge goes
        self.tag = tag  # String tag
        self.remote = remote  # True or False

    def add(self):
        assert self.parent is not self.child, "Trying to create self-loop edge on %s" % self.parent
        if Config().args.verify:
            assert self not in self.parent.outgoing, "Trying to create outgoing edge twice: %s" % self
            assert self not in self.child.incoming, "Trying to create incoming edge twice: %s" % self
            assert self.parent not in self.child.descendants, "Detected cycle created by edge: %s" % self
        self.parent.add_outgoing(self)
        self.child.add_incoming(self)

    def __repr__(self):
        return Edge.__name__ + "(" + self.tag + ", " + repr(self.parent) + ", " + repr(self.child) +\
               ((", " + str(self.remote)) if self.remote else "") + ")"

    def __str__(self):
        return "%s -%s-> %s%s" % (self.parent, self.tag, self.child,
                                  " (remote)" if self.remote else "")

    def __eq__(self, other):
        return other and self.parent.index == other.parent.index and self.child == other.child and \
               self.tag == other.tag and self.remote == other.remote

    def __hash__(self):
        return hash((self.parent.index, self.child.index, self.tag))



"""Testing code for the tupa.model module, unit-testing only."""

import os
import pytest

from tupa.config import CLASSIFIERS
from tupa.model import Model, ClassifierProperty, NODE_LABEL_KEY
from tupa.states.state import State
from .conftest import remove_existing, weight_decay, assert_all_params_equal


def parse(formats, model, passage, train):
    for axis in formats:
        axes = (axis,) + ((NODE_LABEL_KEY,) if axis == "amr" else ())
        model.config.set_format(axis)
        model.init_model(axis)
        state = State(passage)
        if ClassifierProperty.require_init_features in model.classifier_properties:
            model.init_features(state, train=train)
        features = model.feature_extractor.extract_features(state)
        for a in axes:
            pred = model.classifier.score(features, axis=a).argmax()
            if train:
                model.classifier.update(features, axis=a, pred=pred, true=[0])
        model.classifier.finished_step(train=train)
        model.classifier.finished_item(train=train)


@pytest.mark.parametrize("iterations", (1, 2))
@pytest.mark.parametrize("model_type", CLASSIFIERS)
@pytest.mark.parametrize("omit_features", (None, "d"), ids=("", "omitd"))
def test_model(model_type, formats, test_passage, iterations, omit_features, config):
    filename = "_".join(filter(None, [os.path.join("test_files", "models", "test"), model_type, omit_features]+formats))
    remove_existing(filename)
    config.update(dict(classifier=model_type, copy_shared=None, omit_features=omit_features))
    finalized = model = Model(filename, config=config)
    for i in range(iterations):
        parse(formats, model, test_passage, train=True)
        finalized = model.finalize(finished_epoch=True)
        assert not getattr(finalized.feature_extractor, "node_dropout", 0), finalized.feature_extractor.node_dropout
        parse(formats, model, test_passage, train=False)
        finalized.save()
    loaded = Model(filename, config=config)
    loaded.load()
    assert not getattr(loaded.feature_extractor, "node_dropout", 0), loaded.feature_extractor.node_dropout
    for key, param in sorted(model.feature_extractor.params.items()):
        loaded_param = loaded.feature_extractor.params[key]
        assert param == loaded_param
    assert_all_params_equal(finalized.all_params(), loaded.all_params(), decay=weight_decay(model))



"""Testing code for the tupa package, unit-testing only."""

import pytest
from numpy.testing import assert_allclose
from semstr.evaluate import Scores
from ucca import convert

from tupa.config import SPARSE, MLP, BIRNN, HIGHWAY_RNN, NOOP, Iterations
from tupa.parse import Parser, ParserException
from .conftest import FORMATS, remove_existing, passage_files, load_passage, weight_decay, assert_all_params_equal

CLASSIFIERS = (SPARSE, BIRNN, NOOP)


@pytest.mark.parametrize("model_type", CLASSIFIERS)
def test_parser(config, model_type, formats, default_setting, text=True):
    filename = "test_files/models/%s_%s%s" % ("_".join(formats), model_type, default_setting.suffix())
    remove_existing(filename)
    config.update(default_setting.dict())
    scores = []
    params = []
    passages = list(map(load_passage, passage_files(*formats)))
    evaluate = ("amr" not in formats)
    for mode in "train", "load":
        print("-- %sing %s" % (mode, model_type))
        config.update(dict(classifier=model_type, copy_shared=None))
        p = Parser(model_files=filename, config=config)
        p.save_init = True
        list(p.train(passages if mode == "train" else None, dev=passages, test=True, iterations=2))
        assert p.model.is_finalized, "Model should be finalized after %sing" % mode
        assert not getattr(p.model.feature_extractor, "node_dropout", 0), p.model.feature_extractor.node_dropout
        all_params = p.model.all_params()
        params.append(all_params)
        param1, param2 = [d.get("W") for d in (all_params, p.model.feature_extractor.params)]
        if param1 is not None and param2 and param2.init is not None and not config.args.update_word_vectors:
            assert_allclose(param1, weight_decay(p.model) * param2.init, rtol=1e-6)
        text_results = results = list(p.parse(passages, evaluate=evaluate))
        if text:
            print("Converting to text and parsing...")
            text_results = list(p.parse([p3 for p1 in passages for p2 in convert.to_text(p1, sentences=False) for p3
                                         in convert.from_text(p2, p1.ID, extra_format=p1.extra.get("format"))]))
            assert len(results) == len(text_results)
        if evaluate:
            scores.append(Scores(tuple(zip(*results))[1]).average_f1())
            if text:
                for t, (r, s) in zip(text_results, results):
                    print("  %s F1=%.3f" % (r.ID, s.average_f1()))
        assert not list(p.parse(()))  # parsing nothing returns nothing
        print()
    assert_all_params_equal(*params)
    if evaluate:
        print("-- average f1: %.3f, %.3f\n" % tuple(scores))
        assert scores[0] == pytest.approx(scores[1], 0.1)


EXTRA_CLASSIFIERS = (MLP, HIGHWAY_RNN)


@pytest.mark.parametrize("model_type", EXTRA_CLASSIFIERS)
def test_extra_classifiers(config, model_type, default_setting):
    filename = "test_files/models/%s_%s%s" % (FORMATS[0], model_type, default_setting.suffix())
    remove_existing(filename)
    config.update(default_setting.dict())
    passages = list(map(load_passage, passage_files(FORMATS[0])))
    for mode in "train", "load":
        print("-- %sing %s" % (mode, model_type))
        config.update(dict(classifier=model_type, copy_shared=None))
        p = Parser(model_files=filename, config=config)
        list(p.train(passages if mode == "train" else None, dev=passages, test=True, iterations=2))
        assert p.model.is_finalized, "Model should be finalized after %sing" % mode
        assert not getattr(p.model.feature_extractor, "node_dropout", 0), p.model.feature_extractor.node_dropout


@pytest.mark.parametrize("model_type", (BIRNN,))
def test_copy_shared(config, model_type):
    filename = "test_files/models/%s_%s_copy_shared" % ("_".join(FORMATS), model_type)
    remove_existing(filename)
    config.update(dict(classifier=model_type, lstm_layers=0, copy_shared=[FORMATS[0]]))
    for formats in ((FORMATS[0],), FORMATS):
        p = Parser(model_files=filename, config=config)
        passages = list(map(load_passage, passage_files(*formats)))
        list(p.train(passages, dev=passages, test=True, iterations=2))
        config.update_hyperparams(ucca={"lstm_layers": 1})


@pytest.mark.parametrize("model_type", (BIRNN,))
def test_ensemble(config, model_type):
    config.update(dict(classifier=model_type, lstm_layers=0))
    filenames = ["test_files/models/%s_%s_ensemble%d" % (FORMATS[0], model_type, i) for i in range(1, 3)]
    passages = list(map(load_passage, passage_files(FORMATS[0])))
    for i, filename in enumerate(filenames, start=1):
        config.update(dict(seed=i))
        remove_existing(filename)
        list(Parser(model_files=filename, config=config).train(passages, dev=passages, iterations=2))
    list(Parser(model_files=filenames, config=config).parse(passages, evaluate=True))


@pytest.mark.parametrize("model_type", (BIRNN,))
def test_empty_features(empty_features_config, model_type):
    filename = "test_files/models/%s_%s_empty_features" % (FORMATS[0], model_type)
    remove_existing(filename)
    empty_features_config.update(dict(classifier=model_type))
    passages = list(map(load_passage, passage_files(FORMATS[0])))
    p = Parser(model_files=filename, config=empty_features_config)
    list(p.train(passages, dev=passages, test=True, iterations=2))
    list(p.parse(passages, evaluate=True))


@pytest.mark.parametrize("model_type", (NOOP,))
def test_iterations(config, model_type):
    filename = "test_files/models/%s_%s_iterations" % (FORMATS[0], model_type)
    remove_existing(filename)
    config.update(dict(classifier=model_type))
    passages = list(map(load_passage, passage_files(FORMATS[0])))
    last = 0
    iterations = []
    for i in 2, 5, 9, (11, True), (4, True):
        if isinstance(i, tuple):
            i, simple = i
        else:
            simple = False
            iterations.append(Iterations("%d --word-dim=%d" % (i, i)))
        scores = list(Parser(model_files=filename, config=config).train(passages, dev=passages,
                                                                        iterations=i if simple else iterations))
        assert max(0, i - last) == len(scores)
        last = i
    for iterations in ((3, 2), (4, 4)):
        with pytest.raises(ValueError):  # Number of epochs must be strictly increasing
            list(Parser(model_files=filename, config=config).train(passages, dev=passages, iterations=iterations))


@pytest.mark.parametrize("model_type", CLASSIFIERS)
def test_train_empty(config, model_type, default_setting):
    config.update(default_setting.dict())
    config.update(dict(classifier=model_type))
    filename = "empty"
    remove_existing(filename)
    with pytest.raises(ParserException):
        list(Parser(model_files=filename, config=config).train(load_passage("nonexistent file")))


# @pytest.mark.parametrize("model_type", (BIRNN,))
# def test_change_hyperparams_load(empty_features_config, model_type):
#     filename = "test_files/models/%s_%s_change_hyperparams" % (FORMATS[0], model_type)
#     remove_existing(filename)
#     c = empty_features_config.copy()
#     c.update(dict(classifier=model_type))
#     passages = load_passages(FORMATS[0])
#     for _ in range(2):
#         p = Parser(model_files=filename, config=c)
#         list(p.train(passages, dev=passages, test=True, iterations=2))
#         list(p.parse(passages, evaluate=True))
#         c.update({"ner_dim": 1, "action_dim": 1, "word_dim_external": 1, "word_dim": 1, "node_label_dim": 1,
#                   "node_category_dim": 1, "edge_label_dim": 1, "tag_dim": 1, "dep_dim": 1})



from collections import OrderedDict

from .feature_extractor import FeatureExtractor, calc
from .feature_params import FeatureParameters, NumericFeatureParameters
from ..model_util import UNKNOWN_VALUE, MISSING_VALUE, UnknownDict, save_dict, load_dict

FEATURE_TEMPLATES = (
    "s0s1xd" "s1s0x" "s0b0xd" "b0s0x"  # specific edges
    "s0wmtudencpT#^$xhqyPCIEMN"
    "s1wmtudencT#^$xhyN"
    "s2wmtudencT#^$xhy"
    "s3wmtudencT#^$xhyN"
    "b0wmtudncT#^$hPCIEMN"
    "b1wmtudncT#^$"
    "b2wmtudncT#^$"
    "b3wmtudncT#^$"
    "s0lwmenc#^$"
    "s0rwmenc#^$"
    "s1lwmenc#^$"
    "s1rwmenc#^$"
    "s0llwmen#^$"
    "s0lrwmen#^$"
    "s0rlwmen#^$"
    "s0rrwmen#^$"
    "s1llwmen#^$"
    "s1lrwmen#^$"
    "s1rlwmen#^$"
    "s1rrwmen#^$"
    "s0Lwmen#^$"
    "s0Rwmen#^$"
    "s1Lwmen#^$"
    "s1Rwmen#^$"
    "b0Lwmen#^$"
    "b0Rwmen#^$"
    "s0b0e" "b0s0e"  # specific edges
    "a0eAa1eA",  # past actions
)
INDEXED = "wmtudT"  # words, lemmas, fine POS tags, coarse/universal POS tags, dep rels, entity type
DEFAULT = ()  # intermediate value for missing features
FILENAME_SUFFIX = ".enum"


class DenseFeatureExtractor(FeatureExtractor):
    """
    Extracts features from the parser state for classification. To be used with a NeuralNetwork classifier.
    """
    def __init__(self, params, indexed, hierarchical=False, node_dropout=0, init_params=True, omit_features=None):
        super().__init__(feature_templates=FEATURE_TEMPLATES, omit_features=omit_features)
        self.indexed = indexed
        self.hierarchical = hierarchical
        self.node_dropout = node_dropout
        if init_params:
            self.params = OrderedDict((k, p) for k, p in [(NumericFeatureParameters.SUFFIX, NumericFeatureParameters(1))
                                                          ] + list(params.items()))
            for param in self.params.values():
                self.update_param_indexed(param)
            num_values = self.num_values()
            for key, param in self.params.items():
                param.num = num_values[key]
                param.node_dropout = self.node_dropout
        else:
            self.params = params
    
    def init_param(self, key):
        param = self.params[key]
        self.update_param_indexed(param)
        param.num = self.num_values()[key]

    def num_values(self):
        return {k: len(v) for k, v in self.param_values(all_params=True).items()}

    def update_param_indexed(self, param):
        param.indexed = self.indexed and param.prop in INDEXED

    @property
    def feature_template(self):
        return self.feature_templates[0]

    def init_features(self, state):
        features = OrderedDict()
        for key, param in self.params.items():
            if param.indexed and param.enabled:
                values = [calc(n, state, param.prop) for n in state.terminals]
                param.init_data()
                features[key] = [param.data[v] for v in values]
        return features

    def extract_features(self, state):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :return dict of feature name -> list of numeric values
        """
        features = OrderedDict()
        for key, values in self.param_values(state).items():
            param = self.params[key]
            param.init_data()  # Replace categorical values with their values in data dict:
            features[key] = [(UNKNOWN_VALUE if v == DEFAULT else v) if param.numeric else
                             (MISSING_VALUE if v == DEFAULT else (v if param.indexed else param.data[v]))
                             for v in values]
        return features

    def param_values(self, state=None, all_params=False):
        indexed = []
        by_key = OrderedDict()
        by_prop = OrderedDict()
        for key, param in self.params.items():
            if param.enabled and param.dim or all_params:
                if param.indexed:
                    if param.copy_from:
                        copy_from = self.params.get(param.copy_from)
                        if copy_from and copy_from.enabled and copy_from.dim and not all_params:
                            continue
                    if param.prop not in indexed:
                        indexed.append(param.prop)  # Only need one copy of indices
                by_key[key] = by_prop.setdefault(
                    NumericFeatureParameters.SUFFIX if param.numeric else param.prop,
                    ([state.node_ratio()] if state else [1] if all_params else []) if param.numeric else [])
        for e, prop, value in self.feature_template.extract(state, DEFAULT, "".join(indexed), as_tuples=True,
                                                            node_dropout=self.node_dropout,
                                                            hierarchical=self.hierarchical):
            vs = by_prop.get(NumericFeatureParameters.SUFFIX if e.is_numeric(prop) else prop)
            if vs is not None:
                vs.append(value if state else (e, prop))
        return by_key

    def all_features(self):
        return ["".join(self.join_props(vs)) for _, vs in sorted(self.param_values().items(), key=lambda x: x[0])]

    @staticmethod
    def join_props(values):
        prev = None
        ret = []
        for element, prop in values:
            prefix = "" if element.is_numeric(prop) and prev == element.str else element.str
            ret.append(prefix + prop)
            prev = element.str
        return ret

    def finalize(self):
        return type(self)(FeatureParameters.copy(self.params, UnknownDict), self.indexed, init_params=False,
                          omit_features=self.omit_features)

    def unfinalize(self):
        """Undo finalize(): replace each feature parameter's data dict with a DropoutDict again, to keep training"""
        for param in self.params.values():
            param.unfinalize()
            self.node_dropout = param.node_dropout

    def save(self, filename, save_init=True):  # TODO Save to JSON instead of pickle, with data as list (not dict)
        super().save(filename, save_init=save_init)
        save_dict(filename + FILENAME_SUFFIX, FeatureParameters.copy(self.params, copy_init=save_init))

    def load(self, filename, order=None):
        super().load(filename, order)
        self.params = FeatureParameters.copy(load_dict(filename + FILENAME_SUFFIX), UnknownDict, order=order)
        self.node_dropout = 0



from .feature_extractor import FeatureExtractor


class EmptyFeatureExtractor(FeatureExtractor):

    def __init__(self):
        super().__init__()

    def extract_features(self, state):
        return {}



import os
from configargparse import ArgParser

from tupa.config import FORMATS
from tupa.model import NODE_LABEL_KEY
from tupa.scripts.export import load_model


def strip_multitask(model, keep):
    if "amr" not in keep:  # Remove AMR-specific features: node label and category
        delete_if_exists((model.feature_params, model.classifier.params), (NODE_LABEL_KEY, "c"))
    delete_if_exists((model.classifier.labels, model.classifier.axes), {NODE_LABEL_KEY}.union(FORMATS).difference(keep))


def delete_if_exists(dicts, keys):
    for d in dicts:
        for key in keys:
            try:
                del d[key]
            except KeyError:
                pass


def main(args):
    os.makedirs(args.out_dir, exist_ok=True)
    for filename in args.models:
        model = load_model(filename)
        strip_multitask(model, args.keep)
        model.filename = os.path.join(args.out_dir, os.path.basename(filename))
        model.save()


if __name__ == "__main__":
    argparser = ArgParser(description="Load TUPA model and save with just one task's features/weights.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    argparser.add_argument("-k", "--keep", nargs="+", choices=tuple(filter(None, FORMATS)), default=["ucca"],
                           help="tasks to keep features/weights for")
    argparser.add_argument("-o", "--out-dir", default=".", help="directory to write modified model files to")
    main(argparser.parse_args())



"""Testing code for the tupa.features package, unit-testing only."""
import os
from collections import OrderedDict

import pytest
from ucca import textutil

from tupa.action import Actions
from tupa.features.dense_features import DenseFeatureExtractor
from tupa.features.sparse_features import SparseFeatureExtractor
from tupa.model import Model
from tupa.oracle import Oracle
from tupa.states.state import State
from .conftest import passage_files, load_passage, basename

SPARSE = "sparse"
DENSE = "dense"
VOCAB = os.path.join("test_files", "vocab", "en_core_web_lg.csv")
WORD_VECTORS = os.path.join("test_files", "vocab", "wiki.en.vec")
OMITTED = "d"


class FeatureExtractorCreator:
    def __init__(self, name, indexed=False, annotated=False, vocab=None, wordvectors=None, omit=None):
        self.name = name
        self.indexed = indexed
        self.annotated = annotated
        self.vocab = vocab
        self.id = vocab == "-"
        self.wordvectors = wordvectors
        self.omit = omit

    def __str__(self):
        return "-".join([self.name] + [attr for attr in ("indexed", "annotated", "vocab", "id", "wordvectors", "omit")
                                       if getattr(self, attr)])

    def __call__(self, config):
        config.args.vocab = self.vocab
        config.args.word_vectors = self.wordvectors
        config.args.omit_features = self.omit
        return SparseFeatureExtractor(omit_features=self.omit) if self.name == SPARSE else DenseFeatureExtractor(
            OrderedDict((p.name, p.create_from_config()) for p in Model(None, config=config).param_defs()),
            indexed=self.indexed, node_dropout=0, omit_features=self.omit)


def feature_extractors(*args, **kwargs):
    return [FeatureExtractorCreator(SPARSE, *args, **kwargs), FeatureExtractorCreator(DENSE, *args, **kwargs),
            FeatureExtractorCreator(DENSE, *args, indexed=True, **kwargs)]


def extract_features(feature_extractor, state, features):
    values = feature_extractor.extract_features(state)
    if feature_extractor.params:
        for key, vs in values.items():
            assert len(vs) == feature_extractor.params[key].num, key
    features.append(values)


def _test_features(config, feature_extractor_creator, filename, write_features):
    feature_extractor = feature_extractor_creator(config)
    passage = load_passage(filename, annotate=feature_extractor_creator.annotated)
    textutil.annotate(passage, as_array=True, as_extra=False, vocab=config.vocab())
    config.set_format(passage.extra.get("format") or "ucca")
    oracle = Oracle(passage)
    state = State(passage)
    actions = Actions()
    for key, param in feature_extractor.params.items():
        if not param.numeric:
            param.dropout = 0
            feature_extractor.init_param(key)
    features = [feature_extractor.init_features(state)]
    while True:
        extract_features(feature_extractor, state, features)
        action = min(oracle.get_actions(state, actions).values(), key=str)
        state.transition(action)
        if state.need_label:
            extract_features(feature_extractor, state, features)
            label, _ = oracle.get_label(state, action)
            state.label_node(label)
        if state.finished:
            break
    features = ["%s %s\n" % i for f in features if f for i in (sorted(f.items()) + [("", "")])]
    compare_file = os.path.join("test_files", "features", "-".join((basename(filename), str(feature_extractor_creator)))
                                + ".txt")
    if write_features:
        with open(compare_file, "w", encoding="utf-8") as f:
            f.writelines(features)
    with open(compare_file, encoding="utf-8") as f:
        assert f.readlines() == features, compare_file


@pytest.mark.parametrize("feature_extractor_creator",
                         [f for v in (None, "-", VOCAB) for w in (None, WORD_VECTORS) for o in (None, OMITTED)
                          for f in feature_extractors(vocab=v, wordvectors=w, omit=o)], ids=str)
@pytest.mark.parametrize("filename", passage_files(), ids=basename)
def test_features(config, feature_extractor_creator, filename, write_features):
    _test_features(config, feature_extractor_creator, filename, write_features)


@pytest.mark.parametrize("feature_extractor_creator",
                         [f for v in ("-", VOCAB) for w in (None, WORD_VECTORS) for o in (None, OMITTED)
                          for f in feature_extractors(annotated=True, vocab=v, wordvectors=w, omit=o)], ids=str)
@pytest.mark.parametrize("filename", passage_files("conllu"), ids=basename)
def test_features_conllu(config, feature_extractor_creator, filename, write_features):
    _test_features(config, feature_extractor_creator, filename, write_features)


@pytest.mark.parametrize("feature_extractor_creator", [f for o in (None, OMITTED)
                                                       for f in feature_extractors(omit=o)[:-1]], ids=str)
def test_feature_templates(config, feature_extractor_creator, write_features):
    config.set_format("amr")
    feature_extractor = feature_extractor_creator(config)
    features = ["%s\n" % i for i in feature_extractor.all_features()]
    compare_file = os.path.join("test_files", "features", "templates-%s.txt" % str(feature_extractor_creator))
    if write_features:
        with open(compare_file, "w") as f:
            f.writelines(features)
    with open(compare_file) as f:
        assert f.readlines() == features, compare_file



from configargparse import ArgParser

from tupa.scripts.export import load_model


def main():
    argparser = ArgParser(description="Load TUPA model and save again to a different file.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    argparser.add_argument("-s", "--suffix", default=".1", help="filename suffix to append")
    args = argparser.parse_args()
    for filename in args.models:
        model = load_model(filename)
        model.filename += args.suffix
        model.classifier.filename += args.suffix
        model.save()


if __name__ == "__main__":
    main()



from itertools import groupby

import dynet as dy
import re

from .constants import ACTIVATIONS, INITIALIZERS, CategoricalParameter
from .sub_model import SubModel
from .util import randomize_orthonormal

MLP_PARAM_PATTERN = re.compile(r"[Wb]\d+")


class MultilayerPerceptron(SubModel):
    def __init__(self, config, args, model, layers=None, layer_dim=None, output_dim=None, num_labels=None, params=None,
                 **kwargs):
        super().__init__(params=params, **kwargs)
        self.config = config
        self.args = args
        self.model = model
        self.layers = self.args.layers if layers is None else layers
        self.layer_dim = self.args.layer_dim if layer_dim is None else layer_dim
        self.output_dim = self.args.output_dim if output_dim is None else output_dim
        self.activation = CategoricalParameter(ACTIVATIONS, self.args.activation)
        self.init = CategoricalParameter(INITIALIZERS, self.args.init)
        self.dropout = self.args.dropout
        self.gated = 1 if self.args.gated is None else self.args.gated  # None means --gated was given with no argument
        self.num_labels = num_labels
        self.input_dim = self.input_keys = self.weights = None

    @property
    def total_layers(self):
        return self.layers + (1 if self.num_labels else 0)

    def init_params(self, input_dim):
        assert input_dim, "Zero input dimension to MLP"
        self.input_dim = input_dim
        if self.layers > 0:
            W0 = self.params.get("W0")
            if W0:
                value = W0.as_array()
                old_input_dim = value.shape[1]
                extra_dim = self.input_dim - old_input_dim
                if extra_dim:  # Duplicate suffix of input dimension to accommodate extra input
                    assert extra_dim > 0, "%s: input dimension reduced from %d to %d" % (
                        "/".join(self.save_path), old_input_dim, self.input_dim)
                    value[:, -extra_dim:] /= 2
                    W0.set_value(value)
                    extra_value = value[:, -extra_dim:]
                    self.params["W0+"] = extra = self.model.add_parameters(extra_value.shape)
                    extra.set_value(extra_value)
            else:
                hidden_dim = (self.layers - 1) * [self.layer_dim]
                i_dim = [input_dim] + hidden_dim
                o_dim = hidden_dim + [self.output_dim]
                if self.num_labels:  # Adding another layer at the top
                    i_dim.append(self.output_dim)
                    o_dim.append(self.num_labels)
                self.params.update((prefix + str(i), self.model.add_parameters(dims[i], init=self.init()()))
                                   for prefix, dims in (("W", list(zip(o_dim, i_dim))), ("b", o_dim))
                                   for i, dim in enumerate(dims))
                self.verify_dims()
                randomize_orthonormal(*[v for k, v in self.params.items() if MLP_PARAM_PATTERN.match(k)],
                                      activation=self.activation)
                self.config.print("Initializing MLP: %s" % self, level=4)

    def evaluate(self, inputs, train=False):
        """
        Apply all MLP layers to concatenated input
        :param inputs: (key, vector) per feature type
        :param train: are we training now?
        :return: output vector of size self.output_dim
        """
        input_keys, inputs = list(map(list, zip(*list(inputs))))
        if self.input_keys:
            assert input_keys == self.input_keys, "Got:     %s\nBut expected input keys: %s" % (
                self.input_keys_str(self.input_keys), self.input_keys_str(input_keys))
        else:
            self.input_keys = input_keys
        if self.gated:
            gates = self.params.get("gates")
            if gates is None:  # FIXME attention weights should not be just parameters, but based on biaffine product?
                gates = self.params["gates"] = self.model.add_parameters((len(inputs), self.gated),
                                                                         init=dy.UniformInitializer(1))
            input_dims = [i.dim()[0][0] for i in inputs]
            max_dim = max(input_dims)
            x = dy.concatenate_cols([dy.concatenate([i, dy.zeroes(max_dim - d)])  # Pad with zeros to get uniform dim
                                     if d < max_dim else i for i, d in zip(inputs, input_dims)]) * gates
            # Possibly multiple "attention heads" -- concatenate outputs to one vector
            inputs = [dy.reshape(x, (x.dim()[0][0] * x.dim()[0][1],))]
        x = dy.concatenate(inputs)
        assert len(x.dim()[0]) == 1, "Input should be a vector, but has dimension " + str(x.dim()[0])
        dim = x.dim()[0][0]
        if self.input_dim:
            assert dim == self.input_dim, "Input dim mismatch: %d != %d" % (dim, self.input_dim)
        else:
            self.init_params(dim)
        self.config.print(self, level=4)
        if self.total_layers:
            if self.weights is None:
                self.weights = [[self.params[prefix + str(i)] for prefix in ("W", "b")]
                                for i in range(self.total_layers)]
                if self.weights[0][0].dim()[0][1] < dim:  # number of columns in W0
                    self.weights[0][0] = dy.concatenate_cols([self.weights[0][0], self.params["W0+"]])
            for i, (W, b) in enumerate(self.weights):
                self.config.print(lambda: x.npvalue().tolist(), level=4)
                try:
                    if train and self.dropout:
                        x = dy.dropout(x, self.dropout)
                    x = self.activation()(W * x + b)
                except ValueError as e:
                    raise ValueError("Error in evaluating layer %d of %d" % (i + 1, self.total_layers)) from e
        self.config.print(lambda: x.npvalue().tolist(), level=4)
        return x

    def save_sub_model(self, d, *args):
        self.verify_dims()
        values = super().save_sub_model(
            d,
            ("layers", self.layers),
            ("total_layers", self.total_layers),
            ("layer_dim", self.layer_dim),
            ("output_dim", self.output_dim),
            ("activation", str(self.activation)),
            ("init", str(self.init)),
            ("dropout", self.dropout),
            ("gated", self.gated),
            ("num_labels", self.num_labels),
            ("input_dim", self.input_dim),
            ("input_keys", self.input_keys),
        )
        self.config.print("Saving MLP: %s" % self, level=4)
        return values

    def load_sub_model(self, d, *args, **kwargs):
        d = super().load_sub_model(d, *args, **kwargs)
        self.args.layers = self.layers = d["layers"]
        self.args.layer_dim = self.layer_dim = d["layer_dim"]
        self.args.output_dim = self.output_dim = d["output_dim"]
        self.args.activation = self.activation.string = d["activation"]
        self.args.init = self.init.string = d["init"]
        self.args.dropout = self.dropout = d["dropout"]
        self.args.gated = self.gated = d.get("gated", 0)
        self.num_labels = d["num_labels"]
        self.input_dim = d["input_dim"]
        self.input_keys = d.get("input_keys")
        self.verify_dims()
        self.config.print("Loading MLP: %s" % self, level=4)

    def verify_dims(self):
        if self.layers > 0:
            self.verify_dim("input_dim", sum(W.as_array().shape[1] for W in map(self.params.get, ("W0", "W0+")) if W))
            self.verify_dim("output_dim", self.params["W" + str(self.layers - 1)].as_array().shape[0])
            if self.num_labels:
                self.verify_dim("num_labels", self.params["W" + str(self.layers)].as_array().shape[0])
    
    def verify_dim(self, attr, val):
        expected = getattr(self, attr)
        assert val == expected, "%s %s: %d, expected: %d" % ("/".join(self.save_path), attr, val, expected)

    def invalidate_caches(self):
        self.weights = None

    @staticmethod
    def input_keys_str(input_keys):
        return None if input_keys is None else " ".join("%s:%d" % (k, len(list(l))) for k, l in groupby(input_keys))

    def __str__(self):
        return "%s layers: %d, total_layers: %d, layer_dim: %d, output_dim: %d, activation: %s, init: %s, " \
               "dropout: %f, gated: %d, num_labels: %s, input_dim: %d, input_keys: %s, params: %s" % (
                "/".join(self.save_path), self.layers, self.total_layers, self.layer_dim, self.output_dim,
                self.activation, self.init, self.dropout, self.gated, self.num_labels, self.input_dim,
                self.input_keys_str(self.input_keys), list(self.params.keys()))



import concurrent.futures
import os
import sys
import time
from collections import defaultdict
from enum import Enum
from functools import partial
from glob import glob

from semstr.convert import FROM_FORMAT, TO_FORMAT, from_text
from semstr.evaluate import EVALUATORS, Scores
from semstr.util.amr import LABEL_ATTRIB, WIKIFIER
from semstr.validation import validate
from tqdm import tqdm
from ucca import diffutil, ioutil, textutil, layer0, layer1
from ucca.evaluation import LABELED, UNLABELED, EVAL_TYPES, evaluate as evaluate_ucca
from ucca.normalization import normalize

from tupa.__version__ import GIT_VERSION
from tupa.config import Config, Iterations
from tupa.model import Model, NODE_LABEL_KEY, ClassifierProperty
from tupa.oracle import Oracle
from tupa.states.state import State
from tupa.traceutil import set_traceback_listener


class ParserException(Exception):
    pass


class ParseMode(Enum):
    train = 1
    dev = 2
    test = 3


class AbstractParser:
    def __init__(self, config, models, training=False, evaluation=False):
        self.config = config
        self.models = models
        self.training = training
        self.evaluation = evaluation
        self.action_count = self.correct_action_count = self.label_count = self.correct_label_count = \
            self.num_tokens = self.f1 = 0
        self.started = time.time()

    @property
    def model(self):
        return self.models[0]

    @model.setter
    def model(self, model):
        self.models[0] = model

    @property
    def duration(self):
        return (time.time() - self.started) or 1.0

    def tokens_per_second(self):
        return self.num_tokens / self.duration


class PassageParser(AbstractParser):
    """ Parser for a single passage, has a state and optionally an oracle """
    def __init__(self, passage, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.passage = self.out = passage
        self.format = self.passage.extra.get("format") if self.training or self.evaluation else \
            sorted(set.intersection(*map(set, filter(None, (self.model.formats, self.config.args.formats)))) or
                   self.model.formats)[0]
        if self.training and self.config.args.verify:
            errors = list(validate(self.passage))
            assert not errors, errors
        self.in_format = self.format or "ucca"
        self.out_format = "ucca" if self.format in (None, "text") else self.format
        if self.config.args.use_bert and self.config.args.bert_multilingual is not None:
            self.lang = self.passage.attrib.get("lang")
            assert self.lang, "Attribute 'lang' is required per passage when using multilingual BERT"
        else:
            self.lang = self.passage.attrib.get("lang", self.config.args.lang)
        # Used in verify_passage to optionally ignore a mismatch in linkage nodes:
        self.ignore_node = None if self.config.args.linkage else lambda n: n.tag == layer1.NodeTags.Linkage
        self.state_hash_history = set()
        self.state = self.oracle = self.eval_type = None

    def init(self):
        self.config.set_format(self.in_format)
        WIKIFIER.enabled = self.config.args.wikification
        self.state = State(self.passage)
        # Passage is considered labeled if there are any edges or node labels in it
        edges, node_labels = map(any, zip(*[(n.outgoing, n.attrib.get(LABEL_ATTRIB))
                                            for n in self.passage.layer(layer1.LAYER_ID).all]))
        self.oracle = Oracle(self.passage) if self.training or self.config.args.verify or (
                (self.config.args.verbose > 1 or self.config.args.use_gold_node_labels or self.config.args.action_stats)
                and (edges or node_labels)) else None
        for model in self.models:
            model.init_model(self.config.format, lang=self.lang if self.config.args.multilingual else None)
            if ClassifierProperty.require_init_features in model.classifier_properties:
                model.init_features(self.state, self.training)

    def parse(self, display=True, write=False, accuracies=None):
        self.init()
        passage_id = self.passage.ID
        try:
            with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                executor.submit(self.parse_internal).result(self.config.args.timeout)
            status = "(%d tokens/s)" % self.tokens_per_second()
        except ParserException as e:
            if self.training:
                raise
            self.config.log("%s %s: %s" % (self.config.passage_word, passage_id, e))
            status = "(failed)"
        except concurrent.futures.TimeoutError:
            self.config.log("%s %s: timeout (%fs)" % (self.config.passage_word, passage_id, self.config.args.timeout))
            status = "(timeout)"
        return self.finish(status, display=display, write=write, accuracies=accuracies)

    def parse_internal(self):
        """
        Internal method to parse a single passage.
        If training, use oracle to train on given passages. Otherwise just parse with classifier.
        """
        self.config.print("  initial state: %s" % self.state)
        while True:
            if self.config.args.check_loops:
                self.check_loop()
            self.label_node()  # In case root node needs labeling
            true_actions = self.get_true_actions()
            action, predicted_action = self.choose(true_actions)
            self.state.transition(action)
            need_label, label, predicted_label, true_label = self.label_node(action)
            if self.config.args.action_stats:
                try:
                    with open(self.config.args.action_stats, "a") as f:
                        print(",".join(map(str, [predicted_action, action] + list(true_actions.values()))), file=f)
                except OSError:
                    pass
            self.config.print(lambda: "\n".join(["  predicted: %-15s true: %-15s taken: %-15s %s" % (
                predicted_action, "|".join(map(str, true_actions.values())), action, self.state) if self.oracle else
                                          "  action: %-15s %s" % (action, self.state)] + (
                ["  predicted label: %-9s true label: %s" % (predicted_label, true_label) if self.oracle and not
                 self.config.args.use_gold_node_labels else "  label: %s" % label] if need_label else []) + [
                "    " + l for l in self.state.log]))
            if self.state.finished:
                return  # action is Finish (or early update is triggered)

    def get_true_actions(self):
        true_actions = {}
        if self.oracle:
            try:
                true_actions = self.oracle.get_actions(self.state, self.model.actions, create=self.training)
            except (AttributeError, AssertionError) as e:
                if self.training:
                    raise ParserException("Error in getting action from oracle during training") from e
        return true_actions

    def get_true_label(self, node):
        try:
            return self.oracle.get_label(self.state, node) if self.oracle else (None, None)
        except AssertionError as e:
            if self.training:
                raise ParserException("Error in getting label from oracle during training") from e
            return None, None

    def label_node(self, action=None):
        true_label = label = predicted_label = None
        need_label = self.state.need_label  # Label action that requires a choice of label
        if need_label:
            true_label, raw_true_label = self.get_true_label(action or need_label)
            label, predicted_label = self.choose(true_label, NODE_LABEL_KEY, "node label")
            self.state.label_node(raw_true_label if label == true_label else label)
        return need_label, label, predicted_label, true_label

    def choose(self, true, axis=None, name="action"):
        if axis is None:
            axis = self.model.axis
        elif axis == NODE_LABEL_KEY and self.config.args.use_gold_node_labels:
            return true, true
        labels = self.model.classifier.labels[axis]
        if axis == NODE_LABEL_KEY:
            true_keys = (labels[true],) if self.oracle else ()  # Must be before score()
            is_valid = self.state.is_valid_label
        else:
            true_keys = None
            is_valid = self.state.is_valid_action
        scores, features = self.model.score(self.state, axis)
        for model in self.models[1:]:  # Ensemble if given more than one model; align label order and add scores
            label_scores = dict(zip(model.classifier.labels[axis].all, self.model.score(self.state, axis)[0]))
            scores += [label_scores.get(a, 0) for a in labels.all]  # Product of Experts, assuming log(softmax)
        self.config.print(lambda: "  %s scores: %s" % (name, tuple(zip(labels.all, scores))), level=4)
        try:
            label = pred = self.predict(scores, labels.all, is_valid)
        except StopIteration as e:
            raise ParserException("No valid %s available\n%s" % (name, self.oracle.log if self.oracle else "")) from e
        label, is_correct, true_keys, true_values = self.correct(axis, label, pred, scores, true, true_keys)
        if self.training:
            if not (is_correct and ClassifierProperty.update_only_on_error in self.model.classifier_properties):
                assert not self.model.is_finalized, "Updating finalized model"
                self.model.classifier.update(
                    features, axis=axis, true=true_keys, pred=labels[pred] if axis == NODE_LABEL_KEY else pred.id,
                    importance=[self.config.args.swap_importance if a.is_swap else 1 for a in true_values] or None)
            if not is_correct and self.config.args.early_update:
                self.state.finished = True
        for model in self.models:
            model.classifier.finished_step(self.training)
            if axis != NODE_LABEL_KEY:
                model.classifier.transition(label, axis=axis)
        return label, pred

    def correct(self, axis, label, pred, scores, true, true_keys):
        true_values = is_correct = ()
        if axis == NODE_LABEL_KEY:
            if self.oracle:
                is_correct = (label == true)
                if is_correct:
                    self.correct_label_count += 1
                elif self.training:
                    label = true
            self.label_count += 1
        else:  # action
            true_keys, true_values = map(list, zip(*true.items())) if true else (None, None)
            label = true.get(pred.id)
            is_correct = (label is not None)
            if is_correct:
                self.correct_action_count += 1
            else:
                label = true_values[scores[true_keys].argmax()] if self.training else pred
            self.action_count += 1
        return label, is_correct, true_keys, true_values

    @staticmethod
    def predict(scores, values, is_valid=None):
        """
        Choose action/label based on classifier
        Usually the best action/label is valid, so max is enough to choose it in O(n) time
        Otherwise, sorts all the other scores to choose the best valid one in O(n lg n)
        :return: valid action/label with maximum probability according to classifier
        """
        return next(filter(is_valid, (values[i] for i in PassageParser.generate_descending(scores))))

    @staticmethod
    def generate_descending(scores):
        yield scores.argmax()
        yield from scores.argsort()[::-1]  # Contains the max, but otherwise items might be missed (different order)

    def finish(self, status, display=True, write=False, accuracies=None):
        self.model.classifier.finished_item(self.training)
        for model in self.models[1:]:
            model.classifier.finished_item(renew=False)  # So that dynet.renew_cg happens only once
        if not self.training or self.config.args.verify:
            self.out = self.state.create_passage(verify=self.config.args.verify, format=self.out_format)
        if write:
            for out_format in self.config.args.formats or [self.out_format]:
                if self.config.args.normalize and out_format == "ucca":
                    normalize(self.out)
                ioutil.write_passage(self.out, output_format=out_format, binary=out_format == "pickle",
                                     outdir=self.config.args.outdir, prefix=self.config.args.prefix,
                                     converter=get_output_converter(out_format), verbose=self.config.args.verbose,
                                     append=self.config.args.join, basename=self.config.args.join)
        if self.oracle and self.config.args.verify:
            self.verify(self.out, self.passage)
        ret = (self.out,)
        if self.evaluation:
            ret += (self.evaluate(self.evaluation),)
            status = "%-14s %s F1=%.3f" % (status, self.eval_type, self.f1)
        if display:
            self.config.print("%s%.3fs %s" % (self.accuracy_str, self.duration, status), level=1)
        if accuracies is not None:
            accuracies[self.passage.ID] = self.correct_action_count / self.action_count if self.action_count else 0
        return ret

    @property
    def accuracy_str(self):
        if self.oracle and self.action_count:
            accuracy_str = "a=%-14s" % percents_str(self.correct_action_count, self.action_count)
            if self.label_count:
                accuracy_str += " l=%-14s" % percents_str(self.correct_label_count, self.label_count)
            return "%-33s" % accuracy_str
        return ""

    def evaluate(self, mode=ParseMode.test):
        if self.format:
            self.config.print("Converting to %s and evaluating..." % self.format)
        self.eval_type = UNLABELED if self.config.is_unlabeled(self.in_format) else LABELED
        evaluator = EVALUATORS.get(self.format, evaluate_ucca)
        score = evaluator(self.out, self.passage, converter=get_output_converter(self.format),
                          verbose=self.out and self.config.args.verbose > 3,
                          constructions=self.config.args.constructions,
                          eval_types=(self.eval_type,) if mode is ParseMode.dev else (LABELED, UNLABELED))
        self.f1 = average_f1(score, self.eval_type)
        score.lang = self.lang
        return score

    def check_loop(self):
        """
        Check if the current state has already occurred, indicating a loop
        """
        h = hash(self.state)
        assert h not in self.state_hash_history, \
            "\n".join(["Transition loop", self.state.str("\n")] + [self.oracle.str("\n")] if self.oracle else ())
        self.state_hash_history.add(h)

    def verify(self, guessed, ref):
        """
        Compare predicted passage to true passage and raise an exception if they differ
        :param ref: true passage
        :param guessed: predicted passage to compare
        """
        assert ref.equals(guessed, ignore_node=self.ignore_node), \
            "Failed to produce true passage" + (diffutil.diff_passages(ref, guessed) if self.training else "")

    @property
    def num_tokens(self):
        return len(set(self.state.terminals).difference(self.state.buffer))  # To count even incomplete parses

    @num_tokens.setter
    def num_tokens(self, _):
        pass


class BatchParser(AbstractParser):
    """ Parser for a single training iteration or single pass over dev/test passages """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seen_per_format = defaultdict(int)
        self.num_passages = 0

    def parse(self, passages, display=True, write=False, accuracies=None):
        passages, total = generate_and_len(single_to_iter(passages))
        if self.config.args.ignore_case:
            passages = to_lower_case(passages)
        pr_width = len(str(total))
        id_width = 1
        if self.config.args.use_bert:
            passages = filter_passages_for_bert(passages, self.config.args)
        passages = self.add_progress_bar(textutil.annotate_all(
            passages, as_array=True, as_extra=False, lang=self.config.args.lang, verbose=self.config.args.verbose > 2,
            vocab=self.model.config.vocab(lang=self.config.args.lang)), display=display)
        for i, passage in enumerate(passages, start=1):
            parser = PassageParser(passage, self.config, self.models, self.training, self.evaluation)
            if self.config.args.verbose and display:
                progress = "%3d%% %*d/%d" % (i / total * 100, pr_width, i, total) if total and i <= total else "%d" % i
                id_width = max(id_width, len(str(passage.ID)))
                print("%s %2s %-6s %-*s" % (progress, parser.lang, parser.in_format, id_width, passage.ID),
                      end=self.config.line_end)
            else:
                passages.set_description()
                postfix = {parser.lang + " " + parser.in_format: passage.ID}
                if display:
                    postfix["|t/s|"] = self.tokens_per_second()
                    if self.correct_action_count:
                        postfix["|a|"] = percents_str(self.correct_action_count, self.action_count, fraction=False)
                    if self.correct_label_count:
                        postfix["|l|"] = percents_str(self.correct_label_count, self.label_count, fraction=False)
                    if self.evaluation and self.num_passages:
                        postfix["|F1|"] = self.f1 / self.num_passages
                passages.set_postfix(**postfix)
            self.seen_per_format[parser.in_format] += 1
            if self.training and self.config.args.max_training_per_format and \
                    self.seen_per_format[parser.in_format] > self.config.args.max_training_per_format:
                self.config.print("skipped", level=1)
                continue
            assert not (self.training and parser.in_format == "text"), "Cannot train on unannotated plain text"
            yield parser.parse(display=display, write=write, accuracies=accuracies)
            self.update_counts(parser)
        if self.num_passages and display:
            self.summary()

    def add_progress_bar(self, it, total=None, display=True):
        return it if self.config.args.verbose and display else tqdm(
            it, unit=self.config.passages_word, total=total, file=sys.stdout, desc="Initializing")

    def update_counts(self, parser):
        self.correct_action_count += parser.correct_action_count
        self.action_count += parser.action_count
        self.correct_label_count += parser.correct_label_count
        self.label_count += parser.label_count
        self.num_tokens += parser.num_tokens
        self.num_passages += 1
        self.f1 += parser.f1

    def summary(self):
        print("Parsed %d%s" % (self.num_passages, self.config.passages_word))
        if self.correct_action_count:
            accuracy_str = percents_str(self.correct_action_count, self.action_count, "correct actions ")
            if self.label_count:
                accuracy_str += ", " + percents_str(self.correct_label_count, self.label_count, "correct labels ")
            print("Overall %s" % accuracy_str)
        print("Total time: %.3fs (average time/%s: %.3fs, average tokens/s: %d)" % (
            self.duration, self.config.passage_word, self.time_per_passage(),
            self.tokens_per_second()), flush=True)

    def time_per_passage(self):
        return self.duration / self.num_passages


class Parser(AbstractParser):
    """ Main class to implement transition-based UCCA parser """
    def __init__(self, model_files=(), config=None, beam=1):
        super().__init__(config=config or Config(),
                         models=list(map(Model, (model_files,) if isinstance(model_files, str) else
                                         model_files or (config.args.classifier,))))
        self.beam = beam  # Currently unused
        self.best_score = self.dev = self.test = self.iteration = self.epoch = self.batch = None
        self.trained = self.save_init = False
        self.accuracies = {}

    def train(self, passages=None, dev=None, test=None, iterations=1):
        """
        Train parser on given passages
        :param passages: iterable of passages to train on
        :param dev: iterable of passages to tune on
        :param test: iterable of passages that would be tested on after train finished
        :param iterations: iterable of Iterations objects whose i attributes are the number of iterations to perform
        """
        self.trained = True
        self.dev = dev
        self.test = test
        if passages:
            self.init_train()
            iterations = [i if isinstance(i, Iterations) else Iterations(i)
                          for i in (iterations if hasattr(iterations, "__iter__") else (iterations,))]
            if any(i.epochs >= j.epochs for i, j in zip(iterations[:-1], iterations[1:])):
                raise ValueError("Arguments to --iterations must be increasing: " + " ".join(map(str, iterations)))
            self.config.args.iterations = iterations
            end = None
            for self.iteration, it in enumerate(iterations, start=1):
                start = self.model.classifier.epoch + 1 if self.model.classifier else 1
                if end and start < end + 1:
                    print("Dropped %d epochs because best score was on %d" % (end - start + 1, start - 1))
                end = it.epochs + 1
                self.config.update_iteration(it)
                if end < start + 1:
                    print("Skipping %s, already trained %s epochs" % (it, start - 1))
                    continue
                for self.epoch in range(start, end):
                    print("Training epoch %d of %d: " % (self.epoch, end - 1))
                    if self.config.args.curriculum and self.accuracies:
                        print("Sorting passages by previous epoch accuracy...")
                        passages = sorted(passages, key=lambda p: self.accuracies.get(p.ID, 0))
                    else:
                        self.config.random.shuffle(passages)
                    if not sum(1 for _ in self.parse(passages, mode=ParseMode.train)):
                        raise ParserException("Could not train on any passage")
                    yield self.eval_and_save(self.iteration == len(iterations) and self.epoch == end - 1,
                                             finished_epoch=True)
                print("Trained %d epochs" % (end - 1))
                if dev:
                    if self.iteration < len(iterations):
                        if self.model.is_retrainable:
                            self.model.load(is_finalized=False)  # Load best model to prepare for next iteration
                    elif test:
                        self.model.load()  # Load best model to prepare for test
        else:  # No passages to train on, just load model
            for model in self.models:
                model.load()
            self.print_config()

    def init_train(self):
        assert len(self.models) == 1, "Can only train one model at a time"
        if self.model.is_retrainable:
            try:
                self.model.load(is_finalized=False)
            except FileNotFoundError:
                print("not found, starting from untrained model.")
        self.print_config()
        self.best_score = self.model.classifier.best_score if self.model.classifier else 0

    def eval_and_save(self, last=False, finished_epoch=False):
        scores = None
        model = self.model
        # noinspection PyAttributeOutsideInit
        self.model = finalized = model.finalize(finished_epoch=finished_epoch)
        if self.dev:
            if not self.best_score:
                self.save(finalized)
            average_score, scores = self.eval(self.dev, ParseMode.dev, self.config.args.devscores)
            if average_score >= self.best_score:
                print("Better than previous best score (%.3f)" % self.best_score)
                finalized.classifier.best_score = average_score
                if self.best_score:
                    self.save(finalized)
                self.best_score = average_score
                if self.config.args.eval_test and self.test and self.test is not True:  # There are passages to parse
                    self.eval(self.test, ParseMode.test, self.config.args.testscores, display=False)
            else:
                print("Not better than previous best score (%.3f)" % self.best_score)
        elif last or self.config.args.save_every is not None:
            self.save(finalized)
        if not last:
            finalized.restore(model)  # Restore non-finalized model
        return scores

    def save(self, model):
        self.config.save(model.filename)
        model.save(save_init=self.save_init)

    def eval(self, passages, mode, scores_filename, display=True):
        print("Evaluating on %s passages" % mode.name)
        passage_scores = [s for _, s in self.parse(passages, mode=mode, evaluate=True, display=display)]
        scores = Scores(passage_scores)
        average_score = average_f1(scores)
        prefix = ".".join(map(str, [self.iteration, self.epoch] + (
            [self.batch] if self.config.args.save_every else [])))
        if display:
            print("Evaluation %s, average %s F1 score on %s: %.3f%s" % (prefix, get_eval_type(scores), mode.name,
                                                                        average_score, scores.details(average_f1)))
        print_scores(scores, scores_filename, prefix=prefix, prefix_title="iteration")
        return average_score, scores

    def parse(self, passages, mode=ParseMode.test, evaluate=False, display=True, write=False):
        """
        Parse given passages
        :param passages: iterable of passages to parse
        :param mode: ParseMode value.
                     If train, use oracle to train on given passages.
                     Otherwise, just parse with classifier.
        :param evaluate: whether to evaluate parsed passages with respect to given ones.
                           Only possible when given passages are annotated.
        :param display: whether to display information on each parsed passage
        :param write: whether to write output passages to file
        :return: generator of parsed passages (or in train mode, the original ones),
                 or, if evaluation=True, of pairs of (Passage, Scores).
        """
        self.batch = 0
        assert mode in ParseMode, "Invalid parse mode: %s" % mode
        training = (mode is ParseMode.train)
        if not training and not self.trained:
            yield from self.train()  # Try to load model from file
        parser = BatchParser(self.config, self.models, training, mode if mode is ParseMode.dev else evaluate)
        for i, passage in enumerate(parser.parse(passages, display=display, write=write, accuracies=self.accuracies),
                                    start=1):
            if training and self.config.args.save_every and i % self.config.args.save_every == 0:
                self.eval_and_save()
                self.batch += 1
            yield passage

    def print_config(self):
        self.config.print("tupa %s" % (self.model.config if self.model else self.config), level=0)


def train_test(train_passages, dev_passages, test_passages, args, model_suffix=""):
    """
    Train and test parser on given passage
    :param train_passages: passage to train on
    :param dev_passages: passages to evaluate on every iteration
    :param test_passages: passages to test on after training
    :param args: extra argument
    :param model_suffix: string to append to model filename before file extension
    :return: generator of Scores objects: dev scores for each training iteration (if given dev), and finally test scores
    """
    model_files = [base + model_suffix + ext for base, ext in map(os.path.splitext, args.models or (args.classifier,))]
    p = Parser(model_files=model_files, config=Config(), beam=args.beam)
    yield from filter(None, p.train(train_passages, dev=dev_passages, test=test_passages, iterations=args.iterations))
    if test_passages:
        if args.train or args.folds:
            print("Evaluating on test passages")
        passage_scores = []
        evaluate = args.evaluate or train_passages
        for result in p.parse(test_passages, evaluate=evaluate, write=args.write):
            _, *score = result
            passage_scores += score
        if passage_scores:
            scores = Scores(passage_scores)
            if args.verbose <= 1 or len(passage_scores) > 1:
                print("\nAverage %s F1 score on test: %.3f" % (get_eval_type(scores), average_f1(scores)))
                print("Aggregated scores:")
                scores.print()
            print_scores(scores, args.testscores)
            yield scores


def get_output_converter(out_format, default=None):
    converter = TO_FORMAT.get(out_format)
    return partial(converter, wikification=Config().args.wikification,
                   verbose=Config().args.verbose > 2) if converter else default


def percents_str(part, total, infix="", fraction=True):
    ret = "%d%%" % (100 * part / total)
    if fraction:
        ret += " %s(%d/%d)" % (infix, part, total)
    return ret


def print_scores(scores, filename, prefix=None, prefix_title=None):
    if filename:
        print_title = not os.path.exists(filename)
        try:
            with open(filename, "a") as f:
                if print_title:
                    titles = scores.titles(LABELED) + scores.titles(UNLABELED)
                    if prefix_title is not None:
                        titles = [prefix_title] + titles
                    print(",".join(titles), file=f)
                fields = scores.fields(LABELED) + scores.fields(UNLABELED)
                if prefix is not None:
                    fields.insert(0, prefix)
                print(",".join(fields), file=f)
        except OSError:
            pass


def single_to_iter(it):
    return it if hasattr(it, "__iter__") else (it,)  # Single passage given


def generate_and_len(it):
    return it, (len(it) if hasattr(it, "__len__") else None)


def to_lower_case(passages):
    for passage in passages:
        for terminal in passage.layer(layer0.LAYER_ID).all:
            terminal.text = terminal.text.lower()
        yield passage


def average_f1(scores, eval_type=None):
    for e in (eval_type or get_eval_type(scores),) + EVAL_TYPES:
        try:
            return scores.average_f1(e)
        except ValueError:
            pass
    return 0


def get_eval_type(scores):
    return UNLABELED if Config().is_unlabeled(scores.format) else LABELED


# Marks input passages as text so that we don't accidentally train on them
def from_text_format(*args, **kwargs):
    for passage in from_text(*args, **kwargs):
        passage.extra["format"] = "text"
        yield passage


CONVERTERS = {k: partial(c, annotate=True) for k, c in FROM_FORMAT.items()}
CONVERTERS[""] = CONVERTERS["txt"] = from_text_format


def read_passages(args, files):
    expanded = [f for pattern in files for f in sorted(glob(pattern)) or (pattern,)]
    return ioutil.read_files_and_dirs(expanded, sentences=args.sentences, paragraphs=args.paragraphs,
                                      converters=CONVERTERS, lang=Config().args.lang)


def filter_passages_for_bert(passages, args):
    from pytorch_pretrained_bert import BertTokenizer
    is_uncased_model = "uncased" in args.bert_model
    tokenizer = BertTokenizer.from_pretrained(args.bert_model, do_lower_case=is_uncased_model)
    for passage in passages:
        text = [node.text for node in passage.layer(layer0.LAYER_ID).all]
        bert_token = tokenizer.tokenize(" ".join(text))
        length = len(bert_token)
        if length >= (512-2):  # -2 for the special separators words.
            lang = passage.attrib.get("lang")
            print("Passage longer than 512!" +
                  "\nFiltering passage:"
                  "\nLang: " + str(lang) +
                  " Id: " + str(passage.ID) +
                  " Tokens Length: " + str(length) +
                  " Passage Length: " + str(len(text)))
        else:
            yield passage


# noinspection PyTypeChecker,PyStringFormat
def main_generator():
    args = Config().args
    assert args.passages or args.train, "Either passages or --train is required (use -h for help)"
    assert args.models or args.train or args.folds, "Either --model or --train or --folds is required"
    assert not (args.train or args.dev) or not args.folds, "--train and --dev are incompatible with --folds"
    assert args.train or not args.dev, "--dev is only possible together with --train"
    if args.folds:
        fold_scores = []
        all_passages = list(read_passages(args, args.passages))
        assert len(all_passages) >= args.folds, \
            "%d folds are not possible with only %d passages" % (args.folds, len(all_passages))
        Config().random.shuffle(all_passages)
        folds = [all_passages[i::args.folds] for i in range(args.folds)]
        for i in range(args.folds):
            print("Fold %d of %d:" % (i + 1, args.folds))
            dev_passages = folds[i]
            test_passages = folds[(i + 1) % args.folds]
            train_passages = [passage for fold in folds if fold is not dev_passages and fold is not test_passages
                              for passage in fold]
            s = list(train_test(train_passages, dev_passages, test_passages, args, "_%d" % i))
            if s and s[-1] is not None:
                fold_scores.append(s[-1])
        if fold_scores:
            scores = Scores(fold_scores)
            print("Average test F1 score for each fold: " + ", ".join("%.3f" % average_f1(s) for s in fold_scores))
            print("Aggregated scores across folds:\n")
            scores.print()
            yield scores
    else:  # Simple train/dev/test by given arguments
        train_passages, dev_passages, test_passages = [read_passages(args, arg) for arg in
                                                       (args.train, args.dev, args.passages)]
        yield from train_test(train_passages, dev_passages, test_passages, args)


def main():
    print("TUPA version " + GIT_VERSION)
    set_traceback_listener()
    list(main_generator())


if __name__ == "__main__":
    main()



from collections import OrderedDict
from enum import Enum

from ucca import textutil
from ucca.layer0 import Terminal

from .action import Actions
from .classifiers.classifier import Classifier
from .config import Config, SEPARATOR, SPARSE, MLP, BIRNN, HIGHWAY_RNN, HIERARCHICAL_RNN, NOOP
from .features.feature_params import FeatureParameters
from .model_util import UnknownDict, AutoIncrementDict, remove_backup, save_json, load_json


class ParameterDefinition:
    def __init__(self, args, name, attr_to_arg, attr_to_val=None):
        self.args = args
        self.name = name
        self.attr_to_arg = attr_to_arg
        self.attr_to_val = attr_to_val or {}

    @property
    def dim_arg(self):
        return self.attr_to_arg["dim"]

    @property
    def size_arg(self):
        return self.attr_to_arg["size"]

    @property
    def enabled(self):
        return bool(getattr(self.args, self.dim_arg))
    
    @enabled.setter
    def enabled(self, value):
        if value:
            raise ValueError("Can only disable parameter configuration by setting 'enabled' to False")
        setattr(self.args, self.dim_arg, 0)

    @property
    def lang_specific(self):
        return self.attr_to_val.get("lang_specific")

    def create_from_config(self, lang=None):
        kwargs = dict(self.attr_to_val)
        kwargs.update({k: getattr(self.get_args(lang), v) for k, v in self.attr_to_arg.items()})
        return FeatureParameters(self.name, **kwargs)

    def load_to_config(self, params):
        for lang in list(self.all_langs(params)) or [None]:
            param = params.get(self.key(lang))
            self.get_args(lang).update({self.dim_arg: 0, self.size_arg: 0} if param is None else
                                       {v: getattr(param, k) for k, v in self.attr_to_arg.items()})

    def get_args(self, lang):
        return self.args.hyperparams.specific[lang] if lang else self.args

    def all_langs(self, params):
        for key in params:
            param_name, _, lang = key.partition(SEPARATOR)
            if param_name == self.name and lang:
                yield lang

    def key(self, lang=None):
        return SEPARATOR.join(filter(None, (self.name, lang)))

    def __str__(self):
        return self.name

    def __repr__(self):
        return "%s(%s, %s)" % (type(self).__name__, self.name, ", ".join(
            "%s=%s" % i for i in list(self.attr_to_arg.items()) + list(self.attr_to_val.items())))


NODE_LABEL_KEY = "n"


class ClassifierProperty(Enum):
    update_only_on_error = 1
    require_init_features = 2
    trainable_after_saving = 3


CLASSIFIER_PROPERTIES = {
    SPARSE: (ClassifierProperty.update_only_on_error,),
    MLP: (ClassifierProperty.trainable_after_saving,),
    BIRNN: (ClassifierProperty.trainable_after_saving, ClassifierProperty.require_init_features),
    HIGHWAY_RNN: (ClassifierProperty.trainable_after_saving, ClassifierProperty.require_init_features),
    HIERARCHICAL_RNN: (ClassifierProperty.trainable_after_saving, ClassifierProperty.require_init_features),
    NOOP: (ClassifierProperty.trainable_after_saving,),
}

NODE_LABEL_PARAM_DEFS = [
    (NODE_LABEL_KEY, dict(dim="node_label_dim",    size="max_node_labels",    dropout="node_label_dropout",
                          min_count="min_node_label_count"))
]
PARAM_DEFS = [
    ("c",            dict(dim="node_category_dim", size="max_node_categories")),
    ("W",            dict(dim="word_dim_external", size="max_words_external", dropout="word_dropout_external",
                          updated="update_word_vectors", filename="word_vectors", vocab="vocab"), dict(
                                                                                 copy_from="w", lang_specific=True)),
    ("w",            dict(dim="word_dim",       size="max_words",  dropout="word_dropout"),  dict(lang_specific=True)),
    ("m",            dict(dim="lemma_dim",      size="max_lemmas", dropout="lemma_dropout"), dict(lang_specific=True)),
    ("t",            dict(dim="tag_dim",        size="max_tags",   dropout="tag_dropout"),   dict(lang_specific=True)),
    ("u",            dict(dim="pos_dim",        size="max_pos",    dropout="pos_dropout")),
    ("d",            dict(dim="dep_dim",        size="max_deps",   dropout="dep_dropout")),
    ("e",            dict(dim="edge_label_dim", size="max_edge_labels")),
    ("p",            dict(dim="punct_dim",      size="max_puncts")),
    ("A",            dict(dim="action_dim",     size="max_action_types")),
    ("T",            dict(dim="ner_dim",        size="max_ner_types")),
    ("#",            dict(dim="shape_dim",      size="max_shapes"),                          dict(lang_specific=True)),
    ("^",            dict(dim="prefix_dim",     size="max_prefixes"),                        dict(lang_specific=True)),
    ("$",            dict(dim="suffix_dim",     size="max_suffixes"),                        dict(lang_specific=True)),
]


class Model:
    def __init__(self, filename, config=None, *args, **kwargs):
        self.config = config or Config().copy()
        self.filename = filename
        self.feature_extractor = self.classifier = self.axis = self.lang = None
        self.feature_params = OrderedDict()
        self.is_finalized = False
        if args or kwargs:
            self.restore(*args, **kwargs)

    def node_label_param_def(self, args=None):
        return self.param_defs(args, only_node_labels=True)[0]

    def param_defs(self, args=None, only_node_labels=False):
        return [ParameterDefinition(args or self.config.args, n, *k) for n, *k in NODE_LABEL_PARAM_DEFS +
                ([] if only_node_labels else PARAM_DEFS)]

    def init_model(self, axis=None, lang=None, init_params=True):
        self.set_axis(axis, lang)
        labels = self.classifier.labels if self.classifier else OrderedDict()
        if init_params:  # Actually use the config state to initialize the features and hyperparameters, otherwise empty
            for param_def in self.param_defs():  # FIXME save parameters separately per format, not just per language
                for param_lang in (param_def.all_langs(self.feature_params) if self.lang else []) \
                        if param_def.lang_specific and self.config.args.multilingual else [None]:
                    key = param_def.key(param_lang)
                    param = self.feature_params.get(key)
                    enabled = param_def.enabled and (not param_lang or param_lang == self.lang)
                    if param:
                        param.enabled = enabled
                    elif self.is_neural_network and enabled:
                        self.feature_params[key] = param_def.create_from_config(param_lang)
                        self.init_param(key)
            if axis and self.axis not in labels:
                labels[self.axis] = self.init_actions()  # Uses config to determine actions
            if self.config.args.node_labels and not self.config.args.use_gold_node_labels and \
                    NODE_LABEL_KEY not in labels:
                labels[NODE_LABEL_KEY] = self.init_node_labels()  # Updates self.feature_params
        if self.classifier:  # Already initialized
            pass
        elif self.config.args.classifier == SPARSE:
            from .features.sparse_features import SparseFeatureExtractor
            from .classifiers.linear.sparse_perceptron import SparsePerceptron
            self.feature_extractor = SparseFeatureExtractor(omit_features=self.config.args.omit_features)
            self.classifier = SparsePerceptron(self.config, labels)
        elif self.config.args.classifier == NOOP:
            from .features.empty_features import EmptyFeatureExtractor
            from .classifiers.noop import NoOp
            self.feature_extractor = EmptyFeatureExtractor()
            self.classifier = NoOp(self.config, labels)
        elif self.is_neural_network:
            from .features.dense_features import DenseFeatureExtractor
            from .classifiers.nn.neural_network import NeuralNetwork
            self.feature_extractor = DenseFeatureExtractor(self.feature_params,
                                                           indexed=self.config.args.classifier != MLP,
                                                           hierarchical=self.config.args.classifier == HIERARCHICAL_RNN,
                                                           node_dropout=self.config.args.node_dropout,
                                                           omit_features=self.config.args.omit_features)
            self.classifier = NeuralNetwork(self.config, labels)
        else:
            raise ValueError("Invalid model type: '%s'" % self.config.args.classifier)
        self._update_input_params()

    def set_axis(self, axis, lang):
        if axis is not None:
            self.axis = axis
        if self.axis is None:
            self.axis = self.config.format
        if lang is not None:
            self.lang = lang
        if self.lang is not None:
            suffix = SEPARATOR + self.lang
            if not self.axis.endswith(suffix):
                self.axis += suffix

    @property
    def formats(self):
        return [k.partition(SEPARATOR)[0] for k in self.classifier.labels]

    @property
    def is_neural_network(self):
        return self.config.args.classifier in (MLP, BIRNN, HIGHWAY_RNN, HIERARCHICAL_RNN)

    @property
    def is_retrainable(self):
        return ClassifierProperty.trainable_after_saving in self.classifier_properties

    @property
    def classifier_properties(self):
        return CLASSIFIER_PROPERTIES[self.config.args.classifier]

    @property
    def actions(self):
        return self.classifier.labels[self.axis]

    def init_actions(self):
        return Actions(size=self.config.args.max_action_labels)

    def init_param(self, key):
        if self.feature_extractor:
            self.feature_extractor.init_param(key)

    def init_node_labels(self):
        node_labels = self.feature_params.get(NODE_LABEL_KEY)
        if node_labels is None:
            node_labels = self.node_label_param_def().create_from_config()
            if self.is_neural_network:
                self.feature_params[NODE_LABEL_KEY] = node_labels
        self.init_param(NODE_LABEL_KEY)
        node_labels.init_data()
        return node_labels.data

    def score(self, state, axis):
        features = self.feature_extractor.extract_features(state)
        return self.classifier.score(features, axis=axis), features  # scores is a NumPy array

    def init_features(self, state, train):
        self.init_model()
        axes = [self.axis]
        if self.config.args.node_labels and not self.config.args.use_gold_node_labels:
            axes.append(NODE_LABEL_KEY)

        passage = [node.text for node in state.passage.nodes.values() if isinstance(node, Terminal)]
        lang = state.passage.attrib.get("lang")
        self.classifier.init_features(self.feature_extractor.init_features(state), axes, train, passage, lang)

    def finalize(self, finished_epoch):
        """
        Copy model, finalizing features (new values will not be added during subsequent use) and classifier (update it)
        :param finished_epoch: whether this is the end of an epoch (or just intermediate checkpoint), for bookkeeping
        :return: a copy of this model with a new feature extractor and classifier (actually classifier may be the same)
        """
        self.config.print("Finalizing model", level=1)
        self.init_model()
        return Model(None, config=self.config.copy(), model=self, is_finalized=True,
                     feature_extractor=self.feature_extractor.finalize(),
                     classifier=self.classifier.finalize(finished_epoch=finished_epoch))

    def save(self, save_init=False):
        """
        Save feature and classifier parameters to files
        """
        if self.filename is not None:
            self.init_model()
            try:
                self.feature_extractor.save(self.filename, save_init=save_init)
                node_labels = self.feature_extractor.params.get(NODE_LABEL_KEY)
                skip_labels = (NODE_LABEL_KEY,) if node_labels and node_labels.size else ()
                bert_configs = {
                        "use_bert": self.config.args.use_bert,
                        "bert_model": self.config.args.bert_model,
                        "bert_layers": self.config.args.bert_layers,
                        "bert_layers_pooling": self.config.args.bert_layers_pooling,
                        "bert_token_align_by": self.config.args.bert_token_align_by,
                        "bert_multilingual": self.config.args.bert_multilingual,
                        "bert_use_default_word_embeddings": self.config.args.bert_use_default_word_embeddings,
                        "bert_dropout": self.config.args.bert_dropout}\
                    if self.config.args.use_bert else {"use_bert": self.config.args.use_bert}
                self.classifier.save(self.filename, skip_labels=skip_labels,
                                     multilingual=self.config.args.multilingual,
                                     omit_features=self.config.args.omit_features,
                                     **bert_configs)
                textutil.models["vocab"] = self.config.args.vocab
                save_json(self.filename + ".nlp.json", textutil.models)
                remove_backup(self.filename)
            except Exception as e:
                raise IOError("Failed saving model to '%s'" % self.filename) from e

    def load(self, is_finalized=True):
        """
        Load the feature and classifier parameters from files
        :param is_finalized: whether loaded model should be finalized, or allow feature values to be added subsequently
        """
        if self.filename is not None:
            try:
                self.config.args.classifier = Classifier.get_property(self.filename, "type")
                self.config.args.multilingual = Classifier.get_property(self.filename, "multilingual")
                self.config.args.omit_features = Classifier.get_property(self.filename, "omit_features")
                self.config.args.use_bert = Classifier.get_property(self.filename, "use_bert")
                if self.config.args.use_bert:
                    self.config.args.bert_model = Classifier.get_property(self.filename, "bert_model")
                    self.config.args.bert_layers = Classifier.get_property(self.filename, "bert_layers")
                    self.config.args.bert_layers_pooling = Classifier.get_property(self.filename, "bert_layers_pooling")
                    self.config.args.bert_token_align_by = Classifier.get_property(self.filename, "bert_token_align_by")
                    self.config.args.bert_multilingual = Classifier.get_property(self.filename, "bert_multilingual")
                    self.config.args.bert_use_default_word_embeddings =\
                        Classifier.get_property(self.filename, "bert_use_default_word_embeddings")
                    self.config.args.bert_dropout = Classifier.get_property(self.filename, "bert_dropout")
                self.init_model(init_params=False)
                self.feature_extractor.load(self.filename, order=[p.name for p in self.param_defs()])
                if not is_finalized:
                    self.feature_extractor.unfinalize()
                self._update_input_params()  # Must be before classifier.load() because it uses them to init the model
                self.classifier.load(self.filename)
                self.is_finalized = is_finalized
                self.load_labels()
                for param_def in self.param_defs(self.config):
                    param_def.load_to_config(self.feature_extractor.params)
                try:
                    textutil.models.update(load_json(self.filename + ".nlp.json"))
                    vocab = textutil.models.get("vocab")
                    if vocab:
                        self.config.args.vocab = vocab
                except FileNotFoundError:
                    pass
                self.config.print("\n".join("%s: %s" % i for i in self.feature_params.items()), level=1)
            except FileNotFoundError:
                self.feature_extractor = self.classifier = None
                raise
            except Exception as e:
                raise IOError("Failed loading model from '%s'" % self.filename) from e

    def restore(self, model, feature_extractor=None, classifier=None, is_finalized=None):
        """
        Set all attributes to a reference to existing model, except labels, which will be copied.
        :param model: Model to restore
        :param feature_extractor: optional FeatureExtractor to restore instead of model's
        :param classifier: optional Classifier to restore instead of model's
        :param is_finalized: whether the restored model is finalized
        """
        if is_finalized is None:
            is_finalized = model.is_finalized
        self.config.print("Restoring %sfinalized model" % ("" if is_finalized else "non-"), level=1)
        self.filename = model.filename
        self.feature_extractor = feature_extractor or model.feature_extractor
        self.classifier = classifier or model.classifier
        self.is_finalized = is_finalized
        self._update_input_params()
        self.classifier.labels_t = OrderedDict((a, l.save()) for a, l in self.classifier.labels.items())
        self.load_labels()

    def load_labels(self):
        """
        Copy classifier's labels to create new Actions/Labels objects
        Restoring from a model that was just loaded from file, or called by restore()
        """
        for axis, all_size in self.classifier.labels_t.items():  # all_size is a pair of (label list, size limit)
            if axis == NODE_LABEL_KEY:  # These are node labels rather than action labels
                node_labels = self.feature_extractor.params.get(NODE_LABEL_KEY)
                if node_labels and node_labels.size:  # Also used for features, so share the dict
                    del all_size
                    labels = node_labels.data
                else:  # Not used as a feature, just get labels
                    labels = UnknownDict() if self.is_finalized else AutoIncrementDict()
                    labels.load(all_size)
            else:  # Action labels for format determined by axis
                labels = Actions(*all_size)
            self.classifier.labels[axis] = labels

    def _update_input_params(self):
        self.feature_params = self.classifier.input_params = self.feature_extractor.params

    def all_params(self):
        d = OrderedDict()
        d["features"] = self.feature_extractor.all_features()
        d.update(("input_" + k, p.data.all) for k, p in self.feature_extractor.params.items() if p.data)
        d.update(self.classifier.all_params())
        return d



LOSSES = ("softmax", "max_margin")
DEFAULT_LOSS = "softmax"

TRAINERS = {
    "sgd": "SimpleSGDTrainer",
    "cyclic": "CyclicalSGDTrainer",
    "momentum": "MomentumSGDTrainer",
    "adagrad": "AdagradTrainer",
    "adadelta": "AdadeltaTrainer",
    "rmsprop": "RMSPropTrainer",
    "adam": "AdamTrainer",
    "amsgrad": "AmsgradTrainer",
}
DEFAULT_TRAINER = "sgd"
EXTRA_TRAINER = "amsgrad"
TRAINER_LEARNING_RATE_PARAM_NAMES = {k: "learning_rate" for k in TRAINERS}
TRAINER_LEARNING_RATE_PARAM_NAMES.update(cyclic="learning_rate_max")
TRAINER_KWARGS = {}  # "adam": dict(beta_2=0.9)}

INITIALIZERS = {
    # "saxe": "SaxeInitializer",
    "glorot_uniform": "GlorotInitializer",
    "normal": "NormalInitializer",
}
DEFAULT_INITIALIZER = "glorot_uniform"

ACTIVATIONS = {
    "cube": "cube",
    "tanh": "tanh",
    "sigmoid": "logistic",
    "relu": "rectify",
}
DEFAULT_ACTIVATION = "relu"

RNNS = {
    # "simple": "SimpleRNNBuilder",
    "gru": "GRUBuilder",
    "lstm": "LSTMBuilder",
    "vanilla_lstm": "VanillaLSTMBuilder",
    "compact_vanilla_lstm": "CompactVanillaLSTMBuilder",
    "coupled_lstm": "CoupledLSTMBuilder",
    "fast_lstm": "FastLSTMBuilder",
}
DEFAULT_RNN = "lstm"


class CategoricalParameter:
    def __init__(self, values, string):
        self._value = self._string = None
        self.values = values
        self.string = string

    def __call__(self):
        import dynet as dy
        return getattr(dy, self._value)

    @property
    def string(self):
        return self._string

    @string.setter
    def string(self, s):
        self._string = s
        self._value = self.values.get(s)

    def __str__(self):
        return self._string



#!/usr/bin/env python
import os
import sys
from subprocess import run

from setuptools import setup, find_packages
from setuptools.command.install import install as _install

from tupa.__version__ import VERSION

try:
    this_file = __file__
except NameError:
    this_file = sys.argv[0]
os.chdir(os.path.dirname(os.path.abspath(this_file)))

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

try:
    # noinspection PyPackageRequirements
    import pypandoc
    try:
        pypandoc.convert_file("README.md", "rst", outputfile="README.rst")
    except (IOError, ImportError, RuntimeError):
        pass
    long_description = pypandoc.convert_file("README.md", "rst")
except (IOError, ImportError, RuntimeError):
    long_description = ""


class install(_install):
    # noinspection PyBroadException
    def run(self):
        # Install requirements
        self.announce("Installing dependencies...")
        run(["pip", "--no-cache-dir", "install"] + install_requires, check=True)

        # Install actual package
        _install.run(self)


setup(name="TUPA",
      version=VERSION,
      description="Transition-based UCCA Parser",
      long_description=long_description,
      classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
      ],
      author="Daniel Hershcovich",
      author_email="danielh@cs.huji.ac.il",
      url="https://github.com/huji-nlp/tupa",
      setup_requires=["pypandoc"],
      install_requires=install_requires,
      extras_require={"server": open(os.path.join("server", "requirements.txt")).read().splitlines(),
                      "viz": ["scipy", "pillow", "matplotlib"]},
      packages=find_packages(),
      cmdclass={"install": install},
      entry_points={"console_scripts": ["tupa = tupa.__main__:main"]},
      )



"""Testing code for the tupa.oracle module, unit-testing only."""

import pytest

from tupa.action import Actions
from tupa.oracle import Oracle
from tupa.states.state import State
from .conftest import passage_files, Settings, load_passage, basename


@pytest.mark.parametrize("setting", Settings.all(), ids=str)
@pytest.mark.parametrize("filename", passage_files(), ids=basename)
def test_oracle(config, setting, filename, write_oracle_actions):
    passage = load_passage(filename)
    config.update(setting.dict())
    config.set_format(passage.extra.get("format") or "ucca")
    compare_file = "test_files/oracle_actions/%s%s.txt" % (basename(filename), setting.suffix())
    actions_taken = []
    with open(compare_file, "rw"[write_oracle_actions], encoding="utf-8") as f:
        for i, action in enumerate(gen_actions(passage)):
            if write_oracle_actions:
                print(action, file=f)
            else:
                assert action == f.readline().strip(), "Action %d does not match expected (all actions taken: %s)" % (
                    i, ", ".join(actions_taken) or "none")
            actions_taken.append(action)
    assert len(actions_taken) > 2, passage


def gen_actions(passage):
    oracle = Oracle(passage)
    state = State(passage)
    actions = Actions()
    while True:
        action = min(oracle.get_actions(state, actions).values(), key=str)
        state.transition(action)
        s = str(action)
        if state.need_label:
            label, _ = oracle.get_label(state, action)
            state.label_node(label)
            s += " " + str(label)
        yield s
        if state.finished:
            break



import matplotlib
import numpy as np

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from configargparse import ArgParser
import os
from glob import glob

np.seterr("raise")

PRIMARY_F1_COLUMN = 3
REMOTE_F1_COLUMN = 6


def load_scores(basename, div="dev"):
    filename = "%s.%s.csv" % (basename, div)
    print("Loading %s scores from '%s'" % (div, filename))
    offset = 0
    try:
        with open(filename) as f:
            if "iteration" not in f.readline():
                offset = -1
            scores = np.genfromtxt(f, delimiter=",", invalid_raise=False)
    except ValueError as e:
        raise ValueError("Failed reading '%s'" % filename) from e
    try:
        return scores[:, [PRIMARY_F1_COLUMN + offset, REMOTE_F1_COLUMN + offset]]
    except IndexError:
        try:
            return scores[:, PRIMARY_F1_COLUMN + offset]
        except IndexError as e:
            raise ValueError("Failed reading '%s'" % filename) from e


def visualize(scores, filename, div="dev"):
    plt.plot(range(1, 1 + len(scores)), scores)
    plt.xlabel("epochs")
    plt.ylabel("%s f1" % div)
    if len(scores.shape) > 1:
        plt.legend(["primary", "remote"])
    plt.title(filename)
    output_file = "%s.%s.png" % (filename, div)
    plt.savefig(output_file)
    plt.clf()
    print("Saved '%s'." % output_file)


def main():
    argparser = ArgParser(description="Visualize scores of a model over the dev set, saving to .png file.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    args = argparser.parse_args()
    for pattern in args.models:
        for filename in sorted(glob(pattern)) or [pattern]:
            basename, _ = os.path.splitext(filename)
            for div in "dev", "test":
                try:
                    scores = load_scores(basename, div=div)
                except OSError:
                    continue
                visualize(scores, basename, div=div)


if __name__ == "__main__":
    main()



import csv
import os
from collections import OrderedDict

import numpy as np

from tupa import parse, config
from tupa.config import Config, Iterations
from tupa.model import CLASSIFIER_PROPERTIES, ClassifierProperty

MODELS_DIR = "models"
MAX_ITERATIONS = int(os.environ.get("MAX_ITERATIONS", 50))


class Params:
    def __init__(self, params, **hyperparams):
        self.params = params
        self.hyperparams = hyperparams
        if ClassifierProperty.require_init_features not in CLASSIFIER_PROPERTIES[self.params["classifier"]]:
            for p in [self.params] + list(self.hyperparams.values()):
                p["rnn"] = None
                p["lstm_layer_dim"] = p["lstm_layers"] = p["embedding_layer_dim"] = p["embedding_layers"] = 0
        if self.params["swap"] != config.COMPOUND:
            self.params["max_swap"] = 1
        if not self.params["word_dim_external"]:
            self.params["max_words_external"] = self.params["word_dropout_external"] = 0
            self.params["word_vectors"] = self.params["update_word_vectors"] = None
        if not self.params["word_dim"]:
            self.params["max_words"] = self.params["word_dropout"] = 0
        self.params["models"] = ["%s/%s-%d" % (MODELS_DIR, self.params["classifier"], self.params["seed"])]
        self.scores = None
        self.all_params = OrderedDict(list(self.params.items()) +
                                      [(a + "." + k, v) for a, p in self.hyperparams.items() for k, v in p.items()])

    def run(self, out_file):
        assert Config().args.train and (Config().args.passages or Config().args.dev) or \
               Config().args.passages and Config().args.folds, "insufficient parameters given to parser"
        print("Running with %s" % self)
        Config().update(self.params)
        Config().update_hyperparams(**self.hyperparams)
        for i, self.scores in enumerate(parse.main_generator(), start=1):
            print_title = not os.path.exists(out_file)
            with open(out_file, "a") as f:
                if print_title:
                    csv.writer(f).writerow([k for k in self.all_params.keys()] +
                                           ["average_labeled_f1"] + self.scores.titles())
                csv.writer(f).writerow([str(i if n == "iterations" else p) for n, p in self.all_params.items()] +
                                       [str(self.scores.average_f1())] + self.scores.fields())

    def score(self):
        return -float("inf") if self.scores is None else self.scores.average_f1()

    def __str__(self):
        ret = ", ".join("%s: %s" % (k, v) for k, v in self.all_params.items())
        if self.scores is not None:
            ret += ", average labeled f1: %.3f" % self.score()
        return ret


def get_values_based_on_format(values):
    return values if "amr" in Config().args.formats else (0,)


def main():
    if not os.path.exists(MODELS_DIR):
        os.makedirs(MODELS_DIR)
    Config().args.write = False
    if not Config().args.verbose:
        Config().args.verbose = 1
    out_file = os.environ.get("PARAMS_FILE", "params.csv")
    word_vectors_files = [os.environ[f] for f in os.environ if f.startswith("WORD_VECTORS")]
    size = int(os.environ.get("PARAMS_NUM", 30))
    np.random.seed()
    domains = (
        # Parameter name            Shared  Domain of possible values
        ("seed",                    False,  2147483647),  # max value for int
        ("classifier",              False,  [config.BIRNN, config.HIGHWAY_RNN]),
        ("learning_rate",           False,  [None]),
        ("learning_rate_decay",     False,  5 * [0] + [0.01]),
        ("update_word_vectors",     False,  [True, False]),
        ("word_vectors",            False,  [None] + word_vectors_files),
        ("word_dim_external",       False,  [0] + 5 * [300]),
        ("word_dim",                False,  range(150, 251)),
        ("tag_dim",                 False,  range(15, 26)),
        ("dep_dim",                 False,  range(5, 16)),
        ("edge_label_dim",          False,  range(15, 26)),
        ("node_label_dim",          False,  get_values_based_on_format(range(15, 30))),
        ("node_category_dim",       False,  get_values_based_on_format(range(5, 15))),
        ("max_node_categories",     False,  get_values_based_on_format(range(10, 26))),
        ("punct_dim",               False,  3),
        ("action_dim",              False,  range(2, 6)),
        ("ner_dim",                 False,  range(3, 9)),
        ("max_node_labels",         False,  get_values_based_on_format(range(1000, 4001))),
        ("min_node_label_count",    False,  range(1, 101)),
        ("layer_dim",               False,  range(50, 301)),
        ("layers",                  False,  [2]),
        ("lstm_layer_dim",          True,   range(300, 501, 2)),
        ("lstm_layers",             True,   [2]),
        ("embedding_layer_dim",     True,   range(300, 501)),
        ("embedding_layers",        True,   range(1, 3)),
        ("output_dim",              False,  range(40, 101)),
        ("activation",              True,   ["cube", "relu"]),
        ("init",                    True,   ["glorot_uniform"]),
        ("loss",                    False,  [config.DEFAULT_LOSS]),
        ("minibatch_size",          False,  range(50, 201)),
        ("optimizer",               False,  [config.DEFAULT_TRAINER]),
        ("swap_importance",         False,  np.arange(1, 2, step=.1)),
        ("iterations",              False,  [Iterations(MAX_ITERATIONS)]),
        ("word_dropout",            False,  np.arange(.41, step=.01)),
        ("word_dropout_external",   False,  np.arange(.51, step=.01)),
        ("tag_dropout",             False,  np.arange(.41, step=.01)),
        ("dep_dropout",             False,  np.arange(.41, step=.01)),
        ("node_label_dropout",      False,  np.arange(.41, step=.01)),
        ("node_dropout",            False,  np.arange(.41, step=.01)),
        ("dynet_weight_decay",      False,  [0, 1e-5]),
        ("dropout",                 False,  np.arange(.61, step=.01)),
        ("gated",                   False,  range(5)),
        ("require_connected",       False,  [False]),
        ("swap",                    False,  [config.REGULAR, config.COMPOUND]),
        ("max_swap",                False,  range(2, 6)),
        ("max_words",               False,  range(8000, 30001)),
        ("max_words_external",      False,  [None] + list(range(50000, 250000))),
        ("rnn",                     True,   [config.DEFAULT_RNN]),
    )
    params = [Params(p, shared=s) for p, s in zip(*[map(OrderedDict, zip(*[sample(name, domain, size)
                                                                           for name, shared, domain in domains
                                                                           if shared or all_parameters]))
                                                    for all_parameters in (True, False)])]
    print("All parameter combinations to try:")
    print("\n".join(map(str, params)))
    print("Saving results to '%s'" % out_file)
    for param in params:
        param.run(out_file)
        best = max(params, key=Params.score)
        print("Best parameters: %s" % best)


def sample(name, domain, size):
    return [(name, v.item() if hasattr(v, "item") else v) for v in np.random.choice(domain, size)]


if __name__ == "__main__":
    main()
